#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2009-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    tileGet.py
# @author  Michael Behrisch
# @date    2019-12-11

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import math
import os

try:
    # python3
    import urllib.request as urllib
    from urllib.error import HTTPError as urlerror
except ImportError:
    import urllib
    from urllib2 import HTTPError as urlerror

import optparse

import sumolib  # noqa

MERCATOR_RANGE = 256
MAX_TILE_SIZE = 640
MAPQUEST_TYPES = {"roadmap": "map", "satellite": "sat", "hybrid": "hyb", "terrain": "sat"}


def fromLatLonToPoint(lat, lon):
    # inspired by https://stackoverflow.com/questions/12507274/how-to-get-bounds-of-a-google-static-map
    x = lon * MERCATOR_RANGE / 360
    siny = math.sin(math.radians(lat))
    y = 0.5 * math.log((1 + siny) / (1 - siny)) * -MERCATOR_RANGE / (2 * math.pi)
    return x, y


def fromLatLonToTile(lat, lon, zoom):
    # inspired by https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Python
    n = 2.0 ** zoom
    xtile = int((lon + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.asinh(math.tan(math.radians(lat))) / math.pi) / 2.0 * n)
    return xtile, ytile


def fromTileToLatLon(xtile, ytile, zoom):
    n = 2.0 ** zoom
    lon = xtile / n * 360.0 - 180.0
    lat = math.degrees(math.atan(math.sinh(math.pi * (1 - 2 * ytile / n))))
    return lat, lon


def getZoomWidthHeight(south, west, north, east, maxTileSize):
    center = ((north + south) / 2, (east + west) / 2)
    centerPx = fromLatLonToPoint(*center)
    nePx = fromLatLonToPoint(north, east)
    zoom = 20
    width = (nePx[0] - centerPx[0]) * 2**zoom * 2
    height = (centerPx[1] - nePx[1]) * 2**zoom * 2
    while width > maxTileSize or height > maxTileSize:
        zoom -= 1
        width /= 2
        height /= 2
    return center, zoom, width, height


def retrieveMapServerTiles(url, tiles, west, south, east, north, decals, prefix, net, layer, output_dir):
    zoom = 20
    numTiles = tiles + 1
    while numTiles > tiles:
        zoom -= 1
        sx, sy = fromLatLonToTile(north, west, zoom)
        ex, ey = fromLatLonToTile(south, east, zoom)
        numTiles = (ex - sx + 1) * (ey - sy + 1)
    for x in range(sx, ex + 1):
        for y in range(sy, ey + 1):
            request = "%s/%s/%s/%s" % (url, zoom, y, x)
#            print(request)
#            opener = urllib.build_opener()
#            opener.addheaders = [('User-agent', 'Mozilla/5.0')]
#            urllib.install_opener(opener)
            try:
                urllib.urlretrieve(request, "%s%s_%s.jpeg" % (os.path.join(output_dir, prefix), x, y))
                lat, lon = fromTileToLatLon(x, y, zoom)
                upperLeft = net.convertLonLat2XY(lon, lat)
                lat, lon = fromTileToLatLon(x + 0.5, y + 0.5, zoom)
                center = net.convertLonLat2XY(lon, lat)
                print('    <decal file="%s%s_%s.jpeg" centerX="%s" centerY="%s" width="%s" height="%s" layer="%d"/>' %
                      (prefix, x, y, center[0], center[1],
                       2 * (center[0] - upperLeft[0]), 2 * (upperLeft[1] - center[1]), layer), file=decals)
            except urlerror as e:
                print("Tile server returned HTTP response code: " + str(e.code))
                raise ValueError


optParser = optparse.OptionParser()
optParser.add_option("-p", "--prefix", default="tile", help="for output file")
optParser.add_option("-b", "--bbox", help="bounding box to retrieve in geo coordinates west,south,east,north")
optParser.add_option("-t", "--tiles", type="int",
                     default=1, help="maximum number of tiles the output gets split into")
optParser.add_option("-d", "--output-dir", default=".", help="optional output directory (must already exist)")
optParser.add_option("-s", "--decals-file", default="settings.xml", help="name of decals settings file")
optParser.add_option("-l", "--layer", type="int", default=0,
                     help="(int) layer at which the image will appear, default 0")
optParser.add_option("-x", "--polygon", help="calculate bounding box from polygon data in file")
optParser.add_option("-n", "--net", help="get bounding box from net file")
optParser.add_option("-k", "--key", help="API key to use")
optParser.add_option("-m", "--maptype", default="satellite", help="map type (roadmap, satellite, hybrid, terrain)")
optParser.add_option("-u", "--url",
                     default="services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile",
                     help="Download from the given tile server")
# alternatives: open.mapquestapi.com/staticmap/v4/getmap, maps.googleapis.com/maps/api/staticmap


def get(args=None):
    options, _ = optParser.parse_args(args=args)
    if not options.bbox and not options.net and not options.polygon:
        optParser.error("At least one of 'bbox' and 'net' and 'polygon' has to be set.")
    if not options.url.startswith("http"):
        options.url = "https://" + options.url
    bbox = ((0, 0), (0, 0))
    if options.polygon:
        west = 1e400
        south = 1e400
        east = -1e400
        north = -1e400
        for area in sumolib.output.parse_fast(options.polygon, 'poly', ['shape']):
            coordList = [tuple(map(float, x.split(',')))
                         for x in area.shape.split()]
            for point in coordList:
                west = min(point[0], west)
                south = min(point[1], south)
                east = max(point[0], east)
                north = max(point[1], north)
    if options.bbox:
        west, south, east, north = [float(v) for v in options.bbox.split(',')]
        if south > north or west > east:
            optParser.error("Invalid geocoordinates in bbox.")
    net = None
    if options.net:
        net = sumolib.net.readNet(options.net)
        bbox = net.getBBoxXY()
        west, south = net.convertXY2LonLat(*bbox[0])
        east, north = net.convertXY2LonLat(*bbox[1])

    prefix = os.path.join(options.output_dir, options.prefix)
    mapQuest = "mapquest" in options.url
    with open(os.path.join(options.output_dir, options.decals_file), "w") as decals:
        sumolib.xml.writeHeader(decals, root="viewsettings")
        if "MapServer" in options.url:
            retrieveMapServerTiles(options.url, options.tiles, west, south, east, north,
                                   decals, options.prefix, net, options.layer, options.output_dir)
        else:
            b = west
            for i in range(options.tiles):
                e = b + (east - west) / options.tiles
                offset = (bbox[1][0] - bbox[0][0]) / options.tiles
                c, z, w, h = getZoomWidthHeight(south, b, north, e, 2560 if mapQuest else 640)
                if mapQuest:
                    size = "size=%d,%d" % (w, h)
                    maptype = 'imagetype=png&type=' + MAPQUEST_TYPES[options.maptype]
                else:
                    size = "size=%dx%d" % (w, h)
                    maptype = 'maptype=' + options.maptype
                request = ("%s?%s&center=%.6f,%.6f&zoom=%s&%s&key=%s" %
                           (options.url, size, c[0], c[1], z, maptype, options.key))
    #            print(request)
                try:
                    urllib.urlretrieve(request, "%s%s.png" % (prefix, i))
                    print('    <decal file="%s%s.png" centerX="%s" centerY="%s" width="%s" height="%s" layer="%d"/>' %
                          (options.prefix, i, bbox[0][0] + (i + 0.5) * offset, (bbox[0][1] + bbox[1][1]) / 2,
                           offset, bbox[1][1] - bbox[0][1], options.layer), file=decals)
                    b = e
                except urlerror as e:
                    print("Tile server returned HTTP response code: " + str(e.code))
                    raise ValueError
        print("</viewsettings>", file=decals)


if __name__ == "__main__":
    get()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2019-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    check_constants.py
# @author  Jakob Erdmann
# @date    2019-06-08


from __future__ import print_function
from __future__ import absolute_import
from collections import defaultdict
import constants

occ = defaultdict(list)
for c, val in constants.__dict__.items():
    if isinstance(val, int):
        occ[val].append(c)

print("Duplicate constant use:")
for val, clist in occ.items():
    if len(clist) > 1:
        print("%s : %s" % (hex(val), ' '.join(sorted(clist))))

print("unused constants:")
for i in range(257):
    if len(occ[i]) == 0:
        print(hex(i))



#!/usr/bin/env python3

""" Generate the default Amitran OD-matrix from TAZ weights.

    Author: Lara CODECA

    This program and the accompanying materials are made available under the
    terms of the Eclipse Public License 2.0 which is available at
    http://www.eclipse.org/legal/epl-2.0.
"""

import argparse
import csv
import sys

def get_options(cmd_args=None):
    """ Argument Parser """
    parser = argparse.ArgumentParser(
        prog='generateAmitranFromTAZWeights.py', usage='%(prog)s [options]',
        description='Generate the default Amitran OD-matrix from TAZ weights.')
    parser.add_argument('--taz-weights', type=str, dest='taz_file', required=True,
                        help='Weighted TAZ file (CSV).')
    parser.add_argument('--out', type=str, dest='output', required=True,
                        help='OD matrix in Amitran format.')
    parser.add_argument('--density', type=float, dest='density', default=3000.0,
                        help='Average population density in square kilometers.')
    return parser.parse_args(cmd_args)

class AmitranFromTAZWeightsGenerator():
    """ Generate the default Amitran OD-matrix from TAZ weights. """

    def __init__(self, options):
        self._options = options
        self._taz_weights = dict()
        self._odpairs = list()
        self._load_weights_from_csv()
        self._generate_odpairs_from_taz()

    def _load_weights_from_csv(self):
        """ Load the TAZ weight from a CSV file. """
        with open(self._options.taz_file, 'r') as csvfile:
            weightreader = csv.reader(csvfile)
            header = []
            for row in weightreader:
                if not header:
                    header = row
                elif row: # ignoring empty lines
                    self._taz_weights[row[0]] = {
                        header[0]: row[0],
                        header[1]: row[1],
                        header[2]: int(row[2]),
                        header[3]: float(row[3]),
                    }

    def _generate_odpairs_from_taz(self):
        """ Generate all the possible OD pairs. """
        _single_taz = (len(self._taz_weights) == 1)
        for origin, taz_orig in self._taz_weights.items():
            for destination, _ in self._taz_weights.items():
                if origin == destination and not _single_taz:
                    continue
                amount = round(
                    self._options.density * taz_orig['Area'] / 1e6,     # from mq to square kmq
                    0)
                if amount <= 0:
                    continue
                self._odpairs.append({
                    'origin': origin,
                    'destination': destination,
                    'amount': amount,
                })

    AMITRAN_TPL = """<demand xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="https://sumo.dlr.de/xsd/amitran/od.xsd">
    <actorConfig id="0">
        <timeSlice duration="86400000" startTime="0">{odpair}
        </timeSlice>
    </actorConfig>
</demand>
"""

    ODPAIR_TPL = """
            <odPair amount="{amount}" destination="{dest}" origin="{orig}"/>"""

    def save_odmatrix_to_file(self, filename):
        """ Save the OD-matric in Amitran format. """
        print("Creation of {}".format(filename))
        with open(filename, 'w') as outfile:
            list_of_odpairs = ''
            for pair in self._odpairs:
                list_of_odpairs += self.ODPAIR_TPL.format(amount=int(pair['amount']),
                                                          dest=pair['destination'],
                                                          orig=pair['origin'])
            outfile.write(self.AMITRAN_TPL.format(odpair=list_of_odpairs))
        print("{} created.".format(filename))

def main(cmd_args):
    """ Generate the default Amitran OD-matrix from TAZ weights. """
    options = get_options(cmd_args)
    odmatrix = AmitranFromTAZWeightsGenerator(options)
    odmatrix.save_odmatrix_to_file(options.output)
    print('Done.')

if __name__ == "__main__":
    main(sys.argv[1:])



# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2009-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    route.py
# @author  Michael Behrisch
# @date    2013-10-23

from __future__ import print_function
import os
import sys
SUMO_HOME = os.environ.get('SUMO_HOME',
                           os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..'))
sys.path.append(os.path.join(SUMO_HOME, 'tools'))
from sumolib.miscutils import euclidean  # noqa
from sumolib.geomhelper import polygonOffsetWithMinimumDistanceToPoint  # noqa

try:
    basestring
    # Allows isinstance(foo, basestring) to work in Python 3
except NameError:
    basestring = str


def getLength(net, edges):
    """
    Calculates the length of a route including internal edges.
    The input network has to contain internal edges (withInternal needs to be set when parsing).
    The list of edges can either contain edge objects or edge ids as strings.
    If there is no connection between two consecutive edges, length 0 is assumed (no error is thrown).
    If there are multiple connections of different length, the shortest is used.
    """
    if len(edges) == 0:
        return 0
    if isinstance(edges[0], basestring):
        edges = [net.getEdge(e) for e in edges]
    last = edges[0]
    length = last.getLength()
    for e in edges[1:]:
        if net.hasInternal:
            minInternalCost = 1e400
            for c in last.getConnections(e):
                if c.getViaLaneID() != "":
                    minInternalCost = min(minInternalCost, net.getLane(c.getViaLaneID()).getLength())
            if minInternalCost < 1e400:
                length += minInternalCost
        length += e.getLength()
        last = e
    return length


def _getMinPath(paths):
    minDist = 1e400
    minPath = None
    for path, (dist, _) in paths.items():
        if dist < minDist:
            minPath = path
            minDist = dist
    return minPath


def mapTrace(trace, net, delta, verbose=False, airDistFactor=2, fillGaps=0, gapPenalty=-1, debug=False):
    """
    matching a list of 2D positions to consecutive edges in a network.
    The positions are assumed to be dense (i.e. covering each edge of the route) and in the correct order.
    """
    result = ()
    paths = {}
    lastPos = None
    if verbose:
        print("mapping trace with %s points" % len(trace))
    for pos in trace:
        newPaths = {}
        candidates = net.getNeighboringEdges(pos[0], pos[1], delta, not net.hasInternal)
        if debug:
            print("\n\npos:%s, %s" % (pos[0], pos[1]))
            print("candidates:%s\n" % candidates)
        if len(candidates) == 0 and verbose:
            print("Found no candidate edges for %s,%s" % pos)

        for edge, d in candidates:
            base = polygonOffsetWithMinimumDistanceToPoint(pos, edge.getShape())
            if paths:
                advance = euclidean(lastPos, pos)  # should become a vector
                minDist = 1e400
                minPath = None
                for path, (dist, lastBase) in paths.items():
                    if debug:
                        print("*** extending path %s by edge '%s'" % ([e.getID() for e in path], edge.getID()))
                        print("              lastBase: %s, base: %s, advance: %s, old dist: %s, minDist: %s" %
                              (lastBase, base, advance, dist, minDist))
                    if dist < minDist:
                        if edge == path[-1]:
                            baseDiff = lastBase + advance - base
                            extension = ()
                            if debug:
                                print("---------- same edge")
                        else:
                            maxGap = min(airDistFactor * advance + edge.getLength() + path[-1].getLength(), fillGaps)
                            extension, cost = net.getShortestPath(path[-1], edge, maxGap, includeFromToCost=False)
                            if extension is None:
                                airLineDist = euclidean(
                                    path[-1].getToNode().getCoord(),
                                    edge.getFromNode().getCoord())
                                if gapPenalty < 0:
                                    gapPenalty = airDistFactor * advance
                                baseDiff = abs(lastBase + advance -
                                               path[-1].getLength() - base - airLineDist) + gapPenalty
                                extension = (edge,)
                            else:
                                baseDiff = lastBase + advance - (cost - edge.getLength()) - base
                                extension = extension[1:]
                            if debug:
                                print("---------- extension path: %s, cost: %s, baseDiff: %s" %
                                      (extension, cost, baseDiff))
                        dist += baseDiff * baseDiff
                        if dist < minDist:
                            minDist = dist
                            minPath = path + extension
                        if debug:
                            print("*** new dist: %s baseDiff: %s minDist: %s" % (dist, baseDiff, minDist))
                if minPath:
                    newPaths[minPath] = (minDist, base)
            else:
                newPaths[(edge,)] = (d * d, base)
        if not newPaths:
            if paths:
                result += _getMinPath(paths)
        paths = newPaths
        lastPos = pos
    if paths:
        if debug:
            print("**************** result:")
            for i in result + _getMinPath(paths):
                print("path:%s" % i.getID())
        return result + _getMinPath(paths)
    return result



# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    __init__.py
# @author  Daniel Krajzewicz
# @author  Michael Behrisch
# @date    2012-12-04

from __future__ import absolute_import

from . import poi, polygon  # noqa



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2020-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    setup-sumo.py
# @author  Michael Behrisch
# @date    2020-07-28

from skbuild import setup
import version

SUMO_VERSION = version.get_pep440_version()
setup(
    name='eclipse-sumo',
    version=SUMO_VERSION,
    url='https://sumo.dlr.de/',
    download_url='https://sumo.dlr.de/download',
    author='DLR and contributors',
    author_email='sumo@dlr.de',
    license='EPL-2.0',
    description=("A microscopic, multi-modal traffic simulation package"),

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
        'Programming Language :: C++',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='traffic simulation traci sumo',

    packages=['sumo'],
    package_dir={'': 'tools/build'},

    cmake_install_dir='tools/build/sumo',

    entry_points={
        'console_scripts': [
            'activitygen=sumo:activitygen',
            'dfrouter=sumo:dfrouter',
            'duarouter=sumo:duarouter',
            'emissionsDrivingCycle=sumo:emissionsDrivingCycle',
            'emissionsMap=sumo:emissionsMap',
            'jtrrouter=sumo:jtrrouter',
            'marouter=sumo:marouter',
            'netconvert=sumo:netconvert',
            'netedit=sumo:netedit',
            'netgenerate=sumo:netgenerate',
            'od2trips=sumo:od2trips',
            'polyconvert=sumo:polyconvert',
            'sumo=sumo:sumo',
            'sumo-gui=sumo:sumo_gui',
        ]
    },
)



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2014-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    checkStopOrder.py
# @author  Jakob Erdmann
# @date    2020-03-19

"""
Check order of arriving and leaving vehicles at stopping places.
Report if vehicles change their order at a stop (a vehicle arrives later but leaves earlier)
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
from collections import defaultdict

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    import sumolib  # noqa
    from sumolib.miscutils import parseTime, humanReadableTime
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")


def get_options(args=None):
    parser = sumolib.options.ArgumentParser(description="Sample routes to match counts")
    parser.add_argument("-r", "--route-files", dest="routeFiles",
                        help="Input route file")
    parser.add_argument("-H", "--human-readable-time", dest="hrTime", action="store_true", default=False,
                        help="Write time values as hour:minute:second or day:hour:minute:second rathern than seconds")
    parser.add_argument("-p", "--ignore-parking", dest="ignoreParking", action="store_true", default=False,
                        help="Do not report conflicts with parking vehicles")
    parser.add_argument("--until-from-duration", action="store_true", default=False, dest="untilFromDuration",
                        help="Use stop arrival+duration instead of 'until' to compute overtaking")
    parser.add_argument("--stop-table", dest="stopTable",
                        help="Print timetable information for the given busStop")

    options = parser.parse_args(args=args)
    if options.routeFiles:
        options.routeFiles = options.routeFiles.split(',')
    else:
        print("Argument --route-files is mandatory", file=sys.stderr)
        sys.exit()

    return options


def main(options):

    tf = humanReadableTime if options.hrTime else lambda x: x

    stopTimes = defaultdict(list)
    for routefile in options.routeFiles:
        for vehicle in sumolib.xml.parse(routefile, ['vehicle', 'trip'], heterogeneous=True):
            if vehicle.stop is None:
                continue
            for stop in vehicle.stop:
                if stop.parking in ["true", "True", "1"] and options.ignoreParking:
                    continue
                if options.untilFromDuration:
                    if stop.arrival:
                        until = parseTime(stop.arrival) + parseTime(stop.duration)
                    else:
                        print("Cannot compute 'until' for Vehicle %s because 'arrival' is not defined" %
                              vehicle.id, file=sys.stderr)
                else:
                    until = parseTime(stop.until)
                arrival = parseTime(stop.arrival) if stop.arrival else until - parseTime(stop.duration)
                stopTimes[stop.busStop].append((arrival, until, vehicle.id, stop.getAttributeSecure("tripId", "")))

    for stop, times in stopTimes.items():
        times.sort()
        for i, (a, u, v, t) in enumerate(times):
            for a2, u2, v2, t2 in times[i + 1:]:
                if u2 <= u:
                    print("Vehicle %s (%s, %s) overtakes %s (%s, %s) at stop %s" % (
                        v2, tf(a2), tf(u2), v, tf(a), tf(u), stop), file=sys.stderr)

    if options.stopTable:
        if options.stopTable in stopTimes:
            times = stopTimes[options.stopTable]
            print("# busStop: %s" % options.stopTable)
            print("arrival\tuntil\tveh\ttripId")
            for a, u, v, t in sorted(times):
                print("%s\t%s\t%s\t%s" % (tf(a), tf(u), v, t))
        else:
            print("No vehicle stops at busStop '%s' found" % options.stopTable, file=sys.stderr)


if __name__ == "__main__":
    main(get_options())



#!/usr/bin/env python3

""" SUMO Activity-Based Mobility Generator - Activity console_handlerains

    Author: Lara CODECA

    This program and the accompanying materials are made available under the
    terms of the Eclipse Public License 2.0 whiconsole_handler is available at
    http://www.eclipse.org/legal/epl-2.0.
"""

import collections
import logging
import os
from pprint import pformat
import sys

import numpy
from numpy.random import RandomState

from agsrc import sagaexceptions, sumoutils

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    from traci.exceptions import TraCIException
else:
    sys.exit("Please declare environment variable 'SUMO_HOME'")

class Activities():
    """ Generates the activity chains. """

    ## Activity
    Activity = collections.namedtuple(
        'Activity',
        ['activity', 'fromEdge', 'toEdge', 'arrivalPos', 'start', 'duration', 'final'],
        defaults=(None,) * 7)

    def __init__(self, conf, sumo, environment, logger, profiling=False):
        """
        Initialize the synthetic population.
            :param conf: distionary with the configurations
            :param sumo: already initialized SUMO simulation (TraCI or LibSUMO)
            :param profiling=False: enable cProfile
        """
        self._conf = conf
        self._sumo = sumo
        self._env = environment
        self.logger = logger

        self._max_retry_number = 1000
        if 'maxNumTry' in conf:
            self._max_retry_number = conf['maxNumTry']

        self._profiling = profiling

        self._random_generator = RandomState(seed=self._conf['seed'])

    # Activity Locations

    def _stages_define_locations_position(self, person_stages):
        """ Define the position of each location in the activity chain. """
        home_pos = None
        primary_pos = None

        for pos, stage in person_stages.items():
            if 'Home' in stage.activity:
                if not home_pos:
                    home_pos = self._env.get_random_pos_from_edge(stage.toEdge)
                person_stages[pos] = stage._replace(arrivalPos=home_pos)
            elif 'P-' in stage.activity:
                if not primary_pos:
                    primary_pos = self._env.get_random_pos_from_edge(stage.toEdge)
                person_stages[pos] = stage._replace(arrivalPos=primary_pos)
            else:
                ## Secondary activities
                person_stages[pos] = stage._replace(
                    arrivalPos=self._env.get_random_pos_from_edge(stage.toEdge))

        return person_stages

    def _stages_define_main_locations(self, from_area, to_area, mode):
        """ Define a generic Home and Primary activity location.
            The locations must be reachable in some ways.
        """
        ## Mode split:
        _mode, _ptype, _vtype = sumoutils.get_intermodal_mode_parameters(
            mode, self._conf['intermodalOptions']['vehicleAllowedParking'])

        route = None
        from_edge = None
        to_edge = None
        _retry_counter = 0
        while not route and _retry_counter < self._max_retry_number:
            _retry_counter += 1
            ## Origin and Destination Selection
            from_edge, to_edge = self._env.select_pair(from_area, to_area)
            from_allowed = (
                self._env.sumo_network.getEdge(from_edge).allows('pedestrian') and
                self._env.sumo_network.getEdge(from_edge).allows('passenger') and
                self._env.sumo_network.getEdge(
                    from_edge).getLength() > self._conf['minEdgeAllowed'])
            to_allowed = (
                self._env.sumo_network.getEdge(to_edge).allows('pedestrian') and
                self._env.sumo_network.getEdge(to_edge).allows('passenger') and
                self._env.sumo_network.getEdge(to_edge).getLength() > self._conf['minEdgeAllowed'])
            if self._env.valid_pair(from_edge, to_edge) and from_allowed and to_allowed:
                try:
                    route = self._sumo.simulation.findIntermodalRoute(
                        from_edge, to_edge, modes=_mode, pType=_ptype, vType=_vtype)
                    if not sumoutils.is_valid_route(
                            mode, route, self._conf['intermodalOptions']['vehicleAllowedParking']):
                        route = None
                        self.logger.debug(
                            '_stages_define_main_locations: findIntermodalRoute mode unusable.')
                except TraCIException:
                    self.logger.debug('_stages_define_main_locations: findIntermodalRoute FAILED.')
                    route = None
            else:
                self.logger.debug('_stages_define_main_locations: unusable pair of edges.')
        if route:
            return from_edge, to_edge
        raise sagaexceptions.TripGenerationActivityError(
            'Locations for the main activities not found between {} and {} using {}.'.format(
                from_area, to_area, mode))

    def _stages_define_secondary_locations(self, person_stages, home, primary):
        """ Define secondary activity locations. """
        for pos, stage in person_stages.items():
            if  'S-' in stage.activity:
                ## look for what is coming before
                _prec = None
                _pos = pos - 1
                while not _prec and _pos in person_stages:
                    if 'Home' in person_stages[_pos].activity:
                        _prec = 'H'
                    elif 'P-' in person_stages[_pos].activity:
                        _prec = 'P'
                    _pos -= 1

                ## look for what is coming next
                _succ = None
                _pos = pos + 1
                while not _succ and _pos in person_stages:
                    if 'Home' in person_stages[_pos].activity:
                        _succ = 'H'
                    elif 'P-' in person_stages[_pos].activity:
                        _succ = 'P'
                    _pos += 1

                destination = None
                if _prec == 'H' and _succ == 'H':
                    destination = self._random_location_circle(center=home, other=primary)
                elif _prec == 'P' and _succ == 'P':
                    destination = self._random_location_circle(center=primary, other=home)
                elif _prec != _succ:
                    destination = self._random_location_ellipse(home, primary)
                else:
                    raise sagaexceptions.TripGenerationActivityError(
                        'Invalid sequence in the activity chain: {} --> {}'.format(_prec, _succ),
                        person_stages)

                person_stages[pos] = stage._replace(toEdge=destination)
        return person_stages

    def _random_location_circle(self, center, other):
        """ Return a random edge in within a radius (*) from the given center.

            (*) Uses the ellipses defined by the foci center and other,
                and the major axe of 1.30 * distance between the foci.
        """
        length = None
        try:
            length = self._sumo.simulation.findRoute(center, other).length
        except TraCIException:
            raise sagaexceptions.TripGenerationActivityError(
                'No route between {} and {}'.format(center, other))
        major_axe = length * 1.3
        minor_axe = numpy.sqrt(numpy.square(major_axe) - numpy.square(length))
        radius = minor_axe / 2.0

        self.logger.debug('_random_location_circle: %s [%.2f]', center, radius)
        edges = self._env.get_all_neigh_edges(center, radius)
        if not edges:
            raise sagaexceptions.TripGenerationActivityError(
                'No edges from {} with range {}.'.format(center, length))

        ret = self._random_generator.choice(edges)
        edges.remove(ret)
        allowed = (
            self._env.sumo_network.getEdge(ret).allows('pedestrian') and
            self._env.sumo_network.getEdge(ret).allows('passenger') and
            ret != center and ret != other and
            self._env.sumo_network.getEdge(ret).getLength() > self._conf['minEdgeAllowed'])
        while edges and not allowed:
            ret = self._random_generator.choice(edges)
            edges.remove(ret)
            allowed = (
                self._env.sumo_network.getEdge(ret).allows('pedestrian') and
                self._env.sumo_network.getEdge(ret).allows('passenger') and
                ret != center and ret != other and
                self._env.sumo_network.getEdge(ret).getLength() > self._conf['minEdgeAllowed'])

        if not edges:
            raise sagaexceptions.TripGenerationActivityError(
                'No valid edges from {} with range {}.'.format(center, length))
        return ret

    def _random_location_ellipse(self, focus1, focus2):
        """ Return a random edge in within the ellipse defined by the foci,
            and the major axe of 1.30 * distance between the foci.
        """
        length = None
        try:
            length = self._sumo.simulation.findRoute(focus1, focus2).length
            self.logger.debug('_random_location_ellipse: %s --> %s [%.2f]', focus1, focus2, length)
        except TraCIException:
            raise sagaexceptions.TripGenerationActivityError(
                'No route between {} and {}'.format(focus1, focus2))

        major_axe = length * 1.3

        edges = self._env.get_all_neigh_edges(focus1, length)
        while edges:
            edge = self._random_generator.choice(edges)
            edges.remove(edge)
            if edge in (focus1, focus2):
                continue
            allowed = (
                self._env.sumo_network.getEdge(edge).allows('pedestrian') and
                self._env.sumo_network.getEdge(edge).allows('passenger') and
                self._env.sumo_network.getEdge(edge).getLength() > self._conf['minEdgeAllowed'])
            if not allowed:
                continue
            try:
                first = self._sumo.simulation.findRoute(focus1, edge).length
                second = self._sumo.simulation.findRoute(edge, focus2).length
                if first + second <= major_axe:
                    self.logger.debug('_random_location_ellipse: %s --> %s [%.2f]', focus1, edge, first)
                    self.logger.debug(
                        '_random_location_ellipse: %s --> %s [%.2f]', edge, focus2, second)
                    return edge
            except TraCIException:
                pass

        raise sagaexceptions.TripGenerationActivityError(
            "No location available for _random_location_ellipse [{}, {}]".format(focus1, focus2))

    # Chain

    def generate_person_stages(self, from_area, to_area, activity_chain, mode):
        """ Returns the trip for the given activity chain. """

        # Define a generic Home and Primary activity location.
        from_edge, to_edge = self._stages_define_main_locations(from_area, to_area, mode)

        ## Generate preliminary stages for a person
        person_stages = dict()
        for pos, activity in enumerate(activity_chain):
            if activity not in self._conf['activities']:
                raise sagaexceptions.TripGenerationActivityError(
                    'Activity {} is not define in the config file.'.format(activity))
            _start, _duration = self._get_timing_from_activity(activity)
            if pos == 0:
                if activity != 'Home':
                    raise sagaexceptions.TripGenerationActivityError(
                        "Every activity chain MUST start with 'Home', '{}' given.".format(activity))
                ## Beginning
                person_stages[pos] = self.Activity(
                    activity=activity, fromEdge=from_edge, start=_start, duration=_duration)
            elif 'P-' in activity:
                ## This is a primary activity
                person_stages[pos] = self.Activity(
                    activity=activity, toEdge=to_edge, start=_start, duration=_duration)
            elif 'S-' in activity:
                ## This is a secondary activity
                person_stages[pos] = self.Activity(
                    activity=activity, start=_start, duration=_duration)
            elif activity == 'Home':
                ## End of the activity chain.
                person_stages[pos] = self.Activity(
                    activity=activity, toEdge=from_edge, start=_start, duration=_duration)

        if len(person_stages) <= 2:
            raise sagaexceptions.TripGenerationActivityError(
                "Invalid activity chain. (Minimal: H -> P-? -> H)", activity_chain)

        ## Define secondary activity location
        person_stages = self._stages_define_secondary_locations(person_stages, from_edge, to_edge)

        ## Remove the initial 'Home' stage and update the from of the second stage.
        person_stages[1] = person_stages[1]._replace(fromEdge=person_stages[0].fromEdge)
        if person_stages[0].start:
            person_stages[1] = person_stages[1]._replace(start=person_stages[0].stage)
        del person_stages[0]

        ## Fixing the 'from' field with a forward chain
        pos = 2
        while pos in person_stages:
            person_stages[pos] = person_stages[pos]._replace(fromEdge=person_stages[pos-1].toEdge)
            pos += 1

        ## Compute the real starting time for the activity chain based on ETT and durations
        start = self._stages_compute_start_time(person_stages, mode)
        person_stages[1] = person_stages[1]._replace(start=start)

        ## Define the position of each location in the activity chain.
        person_stages = self._stages_define_locations_position(person_stages)

        ## Final location consistency test
        last_edge = person_stages[1].toEdge
        pos = 2
        while pos in person_stages:
            if person_stages[pos].fromEdge != last_edge:
                raise sagaexceptions.TripGenerationActivityError(
                    'Inconsistency in the locations for the chain of activities.',
                    person_stages)
            last_edge = person_stages[pos].toEdge
            pos += 1

        ## Set the final activity to True
        pos = 1
        while pos in person_stages:
            person_stages[pos] = person_stages[pos]._replace(final=False)
            pos += 1
        person_stages[pos-1] = person_stages[pos-1]._replace(final=True)

        return person_stages

    def _stages_compute_start_time(self, person_stages, mode):
        """ Compute the real starting time for the activity chain. """

        ## Mode split:
        _mode, _ptype, _vtype = sumoutils.get_intermodal_mode_parameters(
            mode, self._conf['intermodalOptions']['vehicleAllowedParking'])

        # Find the first 'start' defined.
        pos = 1
        while pos in person_stages:
            if person_stages[pos].start:
                break
            pos += 1

        start = person_stages[pos].start
        while pos in person_stages:
            ett, route = None, None
            try:
                route = self._sumo.simulation.findIntermodalRoute(
                    person_stages[pos].fromEdge, person_stages[pos].toEdge,
                    modes=_mode, pType=_ptype, vType=_vtype)
                ett = sumoutils.ett_from_route(route)
            except TraCIException:
                raise sagaexceptions.TripGenerationRouteError(
                    'No solution foud for stage {} and modes {}.'.format(
                        pformat(person_stages[pos]), mode))
            if pos-1 in person_stages:
                if person_stages[pos-1].duration:
                    ett += person_stages[pos-1].duration
            start -= ett
            pos -= 1
        return start

    def _get_timing_from_activity(self, activity):
        """ Compute start and duration from the activity defined in the config file. """
        start = None
        if self._conf['activities'][activity]['start']:
            start = self._random_generator.normal(
                loc=self._conf['activities'][activity]['start']['m'],
                scale=self._conf['activities'][activity]['start']['s'])
            if start < 0:
                return self._get_timing_from_activity(activity)
        duration = None
        if self._conf['activities'][activity]['duration']:
            duration = self._random_generator.normal(
                loc=self._conf['activities'][activity]['duration']['m'],
                scale=self._conf['activities'][activity]['duration']['s'])
            if duration <= 0:
                return self._get_timing_from_activity(activity)
        return start, duration



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    stopOrder.py
# @author  Jakob Erdmann
# @date    2020-08-25

"""
Compare ordering of vehicle departure at stops based on a route file with until
times (ground truth) and stop-output
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
from collections import defaultdict

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
import sumolib  # noqa
from sumolib.miscutils import parseTime, humanReadableTime  # noqa
from sumolib.xml import parse  # noqa


def get_options(args=None):
    parser = sumolib.options.ArgumentParser(description="Sample routes to match counts")
    parser.add_argument("-r", "--route-file", dest="routeFile",
                        help="Input route file")
    parser.add_argument("-s", "--stop-file", dest="stopFile",
                        help="Input stop-output file")
    parser.add_argument("-H", "--human-readable-time", dest="hrTime", action="store_true", default=False,
                        help="Write time values as hour:minute:second or day:hour:minute:second rathern than seconds")
    parser.add_argument("-v", "--verbose", action="store_true",
                        default=False, help="tell me what you are doing")

    options = parser.parse_args(args=args)
    if options.routeFile is None or options.stopFile is None:
        parser.print_help()
        sys.exit()

    return options


def main(options):

    def formatVehCode(code):
        time, veh = code
        if options.hrTime:
            time = humanReadableTime(time)
        return "%s (plan=%s)" % (veh, time)

    # stop (stoppingPlaceID or (lane, pos)) -> [(depart1, veh1), (depart2, veh2), ...]
    expected_departs = defaultdict(list)
    actual_departs = defaultdict(dict)
    ignored_stops = 0
    parsed_stops = 0

    for vehicle in parse(options.routeFile, ['vehicle', 'trip'], heterogeneous=True):
        if vehicle.stop is not None:
            for stop in vehicle.stop:
                if stop.hasAttribute("until"):
                    if stop.hasAttribute("busStop"):
                        stopCode = stop.busStop
                    else:
                        stopCode = "%s,%s" % (stop.lane, stop.endPos)
                    expected_departs[stopCode].append((parseTime(stop.until), vehicle.id))
                    parsed_stops += 1
                else:
                    ignored_stops += 1

    print("Parsed %s expected stops at %s locations" % (parsed_stops, len(expected_departs)))
    if ignored_stops > 0:
        sys.stderr.write("Ignored %s stops without 'until' attribute\n" % ignored_stops)

    output_stops = 0
    for stop in parse(options.stopFile, "stopinfo", heterogeneous=True):
        if stop.hasAttribute("busStop"):
            stopCode = stop.busStop
        else:
            stopCode = (stop.lane, stop.endPos)
        ended = parseTime(stop.ended)
        until = ended - parseTime(stop.delay)
        actual_departs[stopCode][(until, stop.id)] = ended
        output_stops += 1
    print("Parsed %s actual stops at %s locations" % (output_stops, len(actual_departs)))

    missing = defaultdict(list)
    for stopCode in sorted(expected_departs.keys()):
        vehicles = expected_departs[stopCode]
        if stopCode in actual_departs:
            actual_vehicles = actual_departs[stopCode]
            comparable_expected = []
            comparable_actual = []
            for vehCode in vehicles:
                if vehCode in actual_vehicles:
                    comparable_expected.append(vehCode)
                    comparable_actual.append((actual_vehicles[vehCode], vehCode))  # (ended, (until, vehID))
                else:
                    missing[stopCode].append(vehCode)
            comparable_expected.sort()
            comparable_actual.sort()
            num_unexpected = len(actual_vehicles) - len(comparable_actual)
            if num_unexpected > 0:
                print("Found %s unexpected stops at %s" % (num_unexpected, stopCode))

            # after sorting, discard the 'ended' attribute and only keep vehCode
            comparable_actual2 = [v[1] for v in comparable_actual]

            # find and remove duplicate
            tmp = []
            for vehCode in comparable_expected:
                if len(tmp) != 0:
                    if vehCode != tmp[-1]:
                        tmp.append(vehCode)
                    else:
                        if options.verbose:
                            print("Found duplicate stop at %s for vehicle %s" % (stopCode, vehCode))
                        comparable_actual2.remove(vehCode)
                else:
                    tmp.append(vehCode)

            comparable_expected = tmp

            if options.verbose:
                actual = [(v[0], v[1][1]) for v in comparable_actual]
                print("stop:", stopCode)
                print("   expected:", comparable_expected)
                print("     actual:", actual)

            for i, vehCode in enumerate(comparable_expected):
                j = comparable_actual2.index(vehCode)
                indexInStops = ""
                if options.verbose:
                    indexInStops = "  Index in stops: ex=%s act=%s" % (i, j)

                if i < j:
                    print("At %s vehicle %s comes after %s.%s" % (
                        stopCode, formatVehCode(vehCode),
                        ','.join(map(formatVehCode, comparable_actual2[i:j])),
                        indexInStops
                    ))
                elif j < i:
                    print("At %s vehicle %s comes before %s.%s" % (
                        stopCode, formatVehCode(vehCode),
                        ','.join(map(formatVehCode, comparable_actual2[j:i])),
                        indexInStops
                    ))
                if i != j:
                    # swap to avoid duplicate out-of-order warnings
                    tmp = comparable_actual2[i]
                    comparable_actual2[i] = comparable_actual2[j]
                    comparable_actual2[j] = tmp
        else:
            missing[stopCode] = vehicles

    print("Simulation missed %s stops at %s locations" % (sum(map(len, missing.values())), len(missing)))


if __name__ == "__main__":
    main(get_options())



# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2013-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    gpsdat.py
# @author  Daniel Krajzewicz
# @author  Michael Behrisch
# @date    2013-01-15

"""
This module includes functions for converting SUMO's fcd-output into
csv data files used by dlr-fcd processing chain
"""
from __future__ import print_function
from __future__ import absolute_import
import datetime
import sumolib.output  # noqa

TAXI_STATUS_FREE_FLOW = "70"


def fcd2gpsdat(inpFCD, outSTRM, further):
    date = further["base-date"]
    for timestep in inpFCD:
        if timestep.vehicle:
            # does not work with subseconds
            mtime = str(
                date + datetime.timedelta(seconds=int(float(timestep.time))))
            for v in timestep.vehicle:
                print('%s\t%s\t%s\t%s\t%s\t%.3f' % (v.id, mtime, v.x, v.y,
                                                    TAXI_STATUS_FREE_FLOW, float(v.speed) * 3.6), file=outSTRM)



# -*- coding: utf-8 -*-
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2011-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    _junction.py
# @author  Michael Behrisch
# @author  Mario Krumnow
# @author  Jakob Erdmann
# @date    2011-03-17

from __future__ import absolute_import
from .domain import Domain
from . import constants as tc


class JunctionDomain(Domain):

    def __init__(self):
        Domain.__init__(self, "junction", tc.CMD_GET_JUNCTION_VARIABLE, tc.CMD_SET_JUNCTION_VARIABLE,
                        tc.CMD_SUBSCRIBE_JUNCTION_VARIABLE, tc.RESPONSE_SUBSCRIBE_JUNCTION_VARIABLE,
                        tc.CMD_SUBSCRIBE_JUNCTION_CONTEXT, tc.RESPONSE_SUBSCRIBE_JUNCTION_CONTEXT)

    def getPosition(self, junctionID, includeZ=False):
        """getPosition(string, bool) -> (double, double) or (double, double, double)

        Returns the coordinates of the center of the junction.
        """
        return self._getUniversal(tc.VAR_POSITION3D if includeZ else tc.VAR_POSITION, junctionID)

    def getShape(self, junctionID):
        """getShape(string) -> list((double, double))

        List of 2D positions (cartesian) describing the geometry.
        """
        return self._getUniversal(tc.VAR_SHAPE, junctionID)



#!/usr/bin/env python3

""" Complete Scenario Generator from OSM.

    Author: Lara CODECA

    This program and the accompanying materials are made available under the
    terms of the Eclipse Public License 2.0 which is available at
    http://www.eclipse.org/legal/epl-2.0.
"""

import argparse
import cProfile
import io
import os
import pstats
import shutil
import subprocess
import sys

from xml.etree import ElementTree

import generateParkingAreasFromOSM
import generateTAZBuildingsFromOSM
import generateTaxiStandsFromOSM
import generateAmitranFromTAZWeights
import generateDefaultsActivityGen
import activitygen
import sagaActivityReport

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    import ptlines2flows
    import generateParkingAreaRerouters
    from visualization import plot_summary
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")

def get_options(cmd_args=None):
    """ Argument Parser. """
    parser = argparse.ArgumentParser(
        prog='complete.generator.py', usage='%(prog)s [options]',
        description='Complete scenario generator from OSM to the ActivityGen.')
    parser.add_argument(
        '--osm', type=str, dest='osm_file', required=True,
        help='OSM file.')
    parser.add_argument(
        '--out', type=str, dest='out_dir', required=True,
        help='Directory for all the output files.')
    parser.add_argument(
        '--lefthand', dest='left_hand_traffic', action='store_true',
        help='Generate a left-hand traffic scenario.')
    parser.set_defaults(left_hand_traffic=False)
    parser.add_argument(
        '--population', type=int, dest='population', default=1000,
        help='Number of people plans to generate.')
    parser.add_argument(
        '--taxi-fleet', type=int, dest='taxi_fleet', default=10,
        help='Size of the taxi fleet.')
    parser.add_argument(
        '--density', type=float, dest='density', default=3000.0,
        help='Average population density in square kilometers.')
    parser.add_argument(
        '--single-taz', dest='single_taz', action='store_true',
        help='Ignore administrative boundaries and generate only one TAZ.')
    parser.set_defaults(single_taz=False)
    parser.add_argument(
        '--admin-level', type=int, dest='admin_level', default=None,
        help='Select only the administrative boundaries with the given level and generate'
        ' the associated TAZs.')
    parser.add_argument(
        '--taz-plot', type=str, dest='html_filename', default='',
        help='Plots the TAZs to an HTML file as OSM overlay. (Requires folium)')
    parser.add_argument(
        '--processes', type=int, dest='processes', default=1,
        help='Number of processes spawned (when suported) to generate the scenario.')
    parser.add_argument(
        '--from-step', type=int, dest='from_step', default=0,
        help='For successive iteration of the script, it defines from which step it should start: '
             '[ 0 - Copy default files.] '
             '[ 1 - Run netconvert & polyconvert.] '
             '[ 2 - Run ptlines2flows.py.] '
             '[ 3 - Generate parking areas.] '
             '[ 4 - Generate parking area rerouters.] '
             '[ 5 - Generate taxi stands.] '
             '[ 6 - Generate taxi stands rerouters.] '
             '[ 7 - Extract TAZ from administrative boundaries.] '
             '[ 8 - Generate OD-matrix.] '
             '[ 9 - Generate SUMOActivityGen defaults.] '
             '[10 - Run SUMOActivityGen.] '
             '[11 - Launch SUMO.] '
             '[12 - Report.] ')
    parser.add_argument(
        '--to-step', type=int, dest='to_step', default=12,
        help='For successive iteration of the script, it defines after which step it should stop: '
             '[ 0 - Copy default files.] '
             '[ 1 - Run netconvert & polyconvert.] '
             '[ 2 - Run ptlines2flows.py.] '
             '[ 3 - Generate parking areas.] '
             '[ 4 - Generate parking area rerouters.] '
             '[ 5 - Generate taxi stands.] '
             '[ 6 - Generate taxi stands rerouters.] '
             '[ 7 - Extract TAZ from administrative boundaries.] '
             '[ 8 - Generate OD-matrix.] '
             '[ 9 - Generate SUMOActivityGen defaults.] '
             '[10 - Run SUMOActivityGen.] '
             '[11 - Launch SUMO.] '
             '[12 - Report.] ')
    parser.add_argument(
        '--profiling', dest='profiling', action='store_true',
        help='Enable Python3 cProfile feature.')
    parser.add_argument(
        '--no-profiling', dest='profiling', action='store_false',
        help='Disable Python3 cProfile feature.')
    parser.set_defaults(profiling=False)
    parser.add_argument(
        '--gui', dest='gui', action='store_true',
        help='Enable SUMO GUI')
    parser.add_argument(
        '--no-gui', dest='gui', action='store_false',
        help='Disable SUMO GUI')
    parser.set_defaults(gui=False)
    parser.add_argument(
        '--local-defaults', dest='local_defaults', action='store_true',
        help='Uses the default folder and files defined locally. If not enabled, uses the files '
        'contained in the sumo/tools/contributed/saga folder.')
    parser.set_defaults(local_defaults=False)
    return parser.parse_args(cmd_args)

## netconvert
DEFAULT_NETCONVERT = 'osm.netccfg'
DEFAULT_NET_XML = 'osm.net.xml'
DEFAULT_PT_STOPS_XML = 'osm_stops.add.xml'
DEFAULT_PT_LINES = 'osm_ptlines.xml'
DEFAULT_SIDE_PARKING_XML = 'osm_parking.xml'
DEFAULT_TYPE_FILES = ('{}/data/typemap/osmNetconvert.typ.xml,'
                      '{}/data/typemap/osmNetconvertBicycle.typ.xml,'
                      '{}/data/typemap/osmNetconvertPedestrians.typ.xml'.format(
                          os.environ['SUMO_HOME'], os.environ['SUMO_HOME'],
                          os.environ['SUMO_HOME']))

## ptlines2flows
DEFAULT_PT_FLOWS = 'osm_pt.rou.xml'

## generateParkingAreasFromOSM
DEFAULT_PARKING_AREAS = 'osm_parking_areas.add.xml'

## merged parking files
DEFAULT_COMPLETE_PARKING_XML = 'osm_complete_parking_areas.add.xml'

## generateParkingAreaRerouters
DEFAULT_PARKING_REROUTERS_XML = 'osm_parking_rerouters.add.xml'

## polyconvert
DEFAULT_POLY_XML = 'osm_polygons.add.xml'

## generateTAZBuildingsFromOSM
DEFAULT_TAZ_OUTPUT_XML = 'osm_taz.xml'
DEFAULT_WEIGHT_OUTPUT_CSV = 'osm_taz_weight.csv'
DEFAULT_BUILDINGS_PREFIX = 'buildings/osm_buildings'

## generateAmitranFromTAZWeights
DEFAULT_ODMATRIX_AMITRAN_XML = 'osm_odmatrix_amitran.xml'

## taxi stands files
DEFAULT_TAXI_STANDS_XML = 'osm_taxi_stands.add.xml'

## generateParkingAreaRerouters for taxi stands
DEFAULT_TAXI_STANDS_REROUTERS_XML = 'osm_taxi_rerouters.add.xml'

## generateDefaultsActivityGen
DEAFULT_GENERIC_AG_CONG = 'activitygen.json'
DEAFULT_SPECIFIC_AG_CONG = 'osm_activitygen.json'

## SUMO
DEFAULT_SUMOCFG = 'osm.sumocfg'
DEFAULT_TRIPINFO_FILE = 'output.tripinfo.xml'
DEFAULT_SUMMARY_FILE = 'output.summary.xml'

## SAGA Report
DEFAULT_REPORT_FILE = 'activities_report.json'

def _call_netconvert(filename, lefthand):
    """ Call netconvert using a subprocess. """
    netconvert_options = ['netconvert',
                          '-c', DEFAULT_NETCONVERT,
                          '--osm', filename,
                          '-o', DEFAULT_NET_XML,
                          '--ptstop-output', DEFAULT_PT_STOPS_XML,
                          '--ptline-output', DEFAULT_PT_LINES,
                          '--parking-output', DEFAULT_SIDE_PARKING_XML,
                          '--type-files', DEFAULT_TYPE_FILES]
    if lefthand:
        netconvert_options.append('--lefthand')
    subprocess.check_call(netconvert_options)

def _call_polyconvert(filename):
    """ Call polyconvert using a subprocess. """
    polyconvert_options = ['polyconvert',
                           '--osm', filename,
                           '--net', DEFAULT_NET_XML,
                           '-o', DEFAULT_POLY_XML]
    subprocess.check_call(polyconvert_options)

def _call_pt_lines_to_flows():
    """ Call directly ptlines2flows from sumo/tools. """
    pt_flows_options = ptlines2flows.get_options(['-n', DEFAULT_NET_XML,
                                                  '-e', '86400',
                                                  '-p', '600',
                                                  '--random-begin',
                                                  '--seed', '42',
                                                  '--ptstops', DEFAULT_PT_STOPS_XML,
                                                  '--ptlines', DEFAULT_PT_LINES,
                                                  '-o', DEFAULT_PT_FLOWS,
                                                  '--ignore-errors',
                                                  '--vtype-prefix', 'pt_',
                                                  '--verbose'])
    ptlines2flows.main(pt_flows_options)

def _call_generate_parking_areas_from_osm(filename):
    """ Call directly generateParkingAreasFromOSM from SUMOActivityGen. """
    parking_options = ['--osm', filename, '--net', DEFAULT_NET_XML, '--out', DEFAULT_PARKING_AREAS]
    generateParkingAreasFromOSM.main(parking_options)

def _merge_parking_files(side_parking, parking_areas, complete_parking):
    """ Merge the two additional files containing parkings into one. """

    side_parking_struct = ElementTree.parse(side_parking).getroot()
    parking_areas_struct = ElementTree.parse(parking_areas).getroot()

    for element in parking_areas_struct:
        side_parking_struct.append(element)

    merged_parking = ElementTree.ElementTree(side_parking_struct)
    merged_parking.write(open(complete_parking, 'wb'))

def _call_generate_parking_area_rerouters(processes):
    """ Call directly generateParkingAreaRerouters from sumo/tools. """
    rerouters_options = ['-a', DEFAULT_COMPLETE_PARKING_XML,
                         '-n', DEFAULT_NET_XML,
                         '--max-number-alternatives', '10',
                         '--max-distance-alternatives', '1000.0',
                         '--min-capacity-visibility-true', '50',
                         '--max-distance-visibility-true', '1000.0',
                         '--processes', str(processes),
                         '-o', DEFAULT_PARKING_REROUTERS_XML, '--tqdm']
    generateParkingAreaRerouters.main(rerouters_options)

def _call_generate_taxi_stands_from_osm(filename):
    """ Call directly generateTaxiStandsFromOSM from SUMOActivityGen. """
    stands_options = ['--osm', filename,
                      '--net', DEFAULT_NET_XML,
                      '--out', DEFAULT_TAXI_STANDS_XML]
    generateTaxiStandsFromOSM.main(stands_options)

def _call_generate_parking_area_rerouters_for_stands(processes):
    """ Call directly generateParkingAreaRerouters from sumo/tools. """
    rerouters_options = ['-a', DEFAULT_TAXI_STANDS_XML,
                         '-n', DEFAULT_NET_XML,
                         '--max-number-alternatives', '10',
                         '--max-distance-alternatives', '5000.0',
                         '--min-capacity-visibility-true', '50',
                         '--max-distance-visibility-true', '1000.0',
                         '--processes', str(processes),
                         '-o', DEFAULT_TAXI_STANDS_REROUTERS_XML, '--tqdm']
    generateParkingAreaRerouters.main(rerouters_options)

def _call_generate_taz_buildings_from_osm(filename, single_taz, processes, admin_level, plot):
    """ Call directly generateTAZBuildingsFromOSM from SUMOActivityGen. """
    taz_buildings_options = ['--osm', filename,
                             '--net', DEFAULT_NET_XML,
                             '--taz-output', DEFAULT_TAZ_OUTPUT_XML,
                             '--weight-output', DEFAULT_WEIGHT_OUTPUT_CSV,
                             '--processes', str(processes),
                             '--poly-output', DEFAULT_BUILDINGS_PREFIX]
    if single_taz:
        taz_buildings_options.append('--single-taz')
    if admin_level:
        taz_buildings_options.append('--admin-level')
        taz_buildings_options.append(str(admin_level))
    if plot:
        taz_buildings_options.append('--taz-plot')
        taz_buildings_options.append(plot)
    generateTAZBuildingsFromOSM.main(taz_buildings_options)

def _call_generate_amitran_from_taz_weights(density):
    """ Call directly generateAmitranFromTAZWeights from SUMOActivityGen. """
    odmatrix_options = ['--taz-weights', DEFAULT_WEIGHT_OUTPUT_CSV,
                        '--out', DEFAULT_ODMATRIX_AMITRAN_XML,
                        '--density', str(density)]
    generateAmitranFromTAZWeights.main(odmatrix_options)

def _call_generate_defaults_activitygen(population, taxi_fleet):
    """ Call directly generateDefaultsActivityGen from SUMOActivityGen. """
    default_options = ['--conf', DEAFULT_GENERIC_AG_CONG,
                       '--od-amitran', DEFAULT_ODMATRIX_AMITRAN_XML,
                       '--out', DEAFULT_SPECIFIC_AG_CONG,
                       '--population', str(population),
                       '--taxi-fleet', str(taxi_fleet)]
    generateDefaultsActivityGen.main(default_options)

def _call_activitygen():
    """ Call directly activitygen from SUMOActivityGen. """
    activitygen_options = ['-c', DEAFULT_SPECIFIC_AG_CONG]
    activitygen.main(activitygen_options)

def _add_rou_to_default_sumocfg():
    """ Load the configuration file used by activitygen and add the newly generated routes. """
    route_files = ''
    for item in os.listdir():
        if '.rou.xml' in item:
            route_files += item + ','
    route_files = route_files.strip(',')

    xml_tree = ElementTree.parse(DEFAULT_SUMOCFG).getroot()
    for field in xml_tree.iter('route-files'):
        ## it should be only one, tho.
        field.attrib['value'] = route_files
    new_sumocfg = ElementTree.ElementTree(xml_tree)
    new_sumocfg.write(open(DEFAULT_SUMOCFG, 'wb'))

def _call_sumo(gui=False):
    """ Call SUMO using a subprocess. """
    if gui:
        subprocess.check_call(['sumo-gui', '-c', DEFAULT_SUMOCFG, '--start'])
    else:
        subprocess.check_call(['sumo', '-c', DEFAULT_SUMOCFG])

def _call_saga_activity_report():
    """ Call directly sagaActivityReport from SUMOActivityGen.. """
    report_options = ['--tripinfo', DEFAULT_TRIPINFO_FILE,
                      '--out', DEFAULT_REPORT_FILE]
    sagaActivityReport.main(report_options)

def _call_plot_summary():
    """ Call directly plot_summary from sumo/tools """
    plot_options = ['-i', DEFAULT_SUMMARY_FILE,
                    '-o', 'summary.png',
                    '--xtime1', '--verbose', '--blind']
    plot_summary.main(plot_options)

def main(cmd_args):
    """ Complete Scenario Generator. """

    args = get_options(cmd_args)
    print(args)

    ## ========================              PROFILER              ======================== ##
    if args.profiling:
        profiler = cProfile.Profile()
        profiler.enable()
    ## ========================              PROFILER              ======================== ##

    os.makedirs(args.out_dir, exist_ok=True)

    if args.from_step <= 0 and args.to_step >= 0:
        print('Copying default configuration files to destination.')
        shutil.copy(args.osm_file, args.out_dir)
        if args.local_defaults:
            shutil.copy('defaults/activitygen.json', args.out_dir)
            shutil.copy('defaults/basic.vType.xml', args.out_dir)
            shutil.copy('defaults/default-gui.xml', args.out_dir)
            shutil.copy('defaults/duarouter.sumocfg', args.out_dir)
            shutil.copy('defaults/osm.netccfg', args.out_dir)
            shutil.copy('defaults/osm.sumocfg', args.out_dir)
        else:
            shutil.copy(
                '{}/contributed/saga/defaults/activitygen.json'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)
            shutil.copy(
                '{}/contributed/saga/defaults/basic.vType.xml'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)
            shutil.copy(
                '{}/contributed/saga/defaults/default-gui.xml'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)
            shutil.copy(
                '{}/contributed/saga/defaults/duarouter.sumocfg'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)
            shutil.copy(
                '{}/contributed/saga/defaults/osm.netccfg'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)
            shutil.copy(
                '{}/contributed/saga/defaults/osm.sumocfg'.format(
                    os.path.join(os.environ['SUMO_HOME'], 'tools')),
                args.out_dir)

    args.osm_file = os.path.basename(args.osm_file)
    os.chdir(args.out_dir)

    if args.from_step <= 1 and args.to_step >= 1:
        print('Generate the net.xml with all the additional components '
                     '(public transports, parkings, ..)')
        _call_netconvert(args.osm_file, args.left_hand_traffic)
        print('Generate polygons using polyconvert.')
        _call_polyconvert(args.osm_file)

    if args.from_step <= 2 and args.to_step >= 2:
        print('Generate flows for public transportation using ptlines2flows.')
        _call_pt_lines_to_flows()

    if args.from_step <= 3 and args.to_step >= 3:
        print('Generate parking area location and possibly merge it with the one provided '
              'by netconvert.')
        _call_generate_parking_areas_from_osm(args.osm_file)
        _merge_parking_files(DEFAULT_SIDE_PARKING_XML, DEFAULT_PARKING_AREAS,
                             DEFAULT_COMPLETE_PARKING_XML)

    if args.from_step <= 4 and args.to_step >= 4:
        print('Generate parking area rerouters using tools/generateParkingAreaRerouters.py')
        _call_generate_parking_area_rerouters(args.processes)

    if args.from_step <= 5 and args.to_step >= 5:
        print('Extract taxi stands from OpenStreetMap.')
        _call_generate_taxi_stands_from_osm(args.osm_file)

    if args.from_step <= 6 and args.to_step >= 6:
        print('Generate taxi stands rerouters using tools/generateParkingAreaRerouters.py')
        _call_generate_parking_area_rerouters_for_stands(args.processes)

    if args.from_step <= 7 and args.to_step >= 7:
        print('Generate TAZ from administrative boundaries, TAZ weights using buildings and '
              'PoIs and the buildings infrastructure.')
        os.makedirs('buildings', exist_ok=True)
        _call_generate_taz_buildings_from_osm(args.osm_file, args.single_taz, args.processes,
                                              args.admin_level, args.html_filename)

    if args.from_step <= 8 and args.to_step >= 8:
        print('Generate the default OD-Matrix in Amitran format. ')
        _call_generate_amitran_from_taz_weights(args.density)

    if args.from_step <= 9 and args.to_step >= 9:
        print('Generate the default values for the activity based mobility generator. ')
        _call_generate_defaults_activitygen(args.population, args.taxi_fleet)

    if args.from_step <= 10 and args.to_step >= 10:
        print('Mobility generation using SUMOActivityGen.')
        _call_activitygen()
        _add_rou_to_default_sumocfg()

    if args.from_step <= 11 and args.to_step >= 11:
        print('Launch sumo.')
        _call_sumo(args.gui)

    if args.from_step <= 12 and args.to_step >= 12:
        print('Report.')
        _call_saga_activity_report()
        _call_plot_summary()

    ## ========================              PROFILER              ======================== ##
    if args.profiling:
        profiler.disable()
        results = io.StringIO()
        pstats.Stats(profiler, stream=results).sort_stats('cumulative').print_stats(25)
        print(results.getvalue())
    ## ========================              PROFILER              ======================== ##

    print('Done.')

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2017-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    setup-libsumo.py
# @author  Benjamin Striner
# @author  Michael Behrisch
# @date    2017-01-26


from setuptools import setup
from setuptools.dist import Distribution
from setuptools.command.install import install
import os
import glob

import version

SUMO_VERSION = version.get_pep440_version()
package_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
data_files = []
for f in sorted(glob.glob(os.path.join(os.path.dirname(package_dir), 'bin', '*.dll'))):
    f = f.lower()
    if not f.endswith("d.dll") or f[:-5] + ".dll" not in data_files:
        data_files.append(f)


class InstallPlatlib(install):
    def finalize_options(self):
        install.finalize_options(self)
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""
    def has_ext_modules(self):
        return True


setup(
    name='libsumo',
    version=SUMO_VERSION,
    url='https://sumo.dlr.de/docs/Libsumo.html',
    author='DLR and contributors',
    author_email='sumo@dlr.de',
    license='EPL-2.0',
    description="The python version of the libsumo API to communicate with the traffic simulation SUMO",

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
        'Programming Language :: Python :: 3',
    ],
    keywords='traffic simulation traci sumo',

    packages=['libsumo'],
    package_dir={'': package_dir},
    package_data={'libsumo': ['*.pyd', '*.so', '*.dylib']},
    data_files=[("", data_files)],
    install_requires=['traci>='+SUMO_VERSION],
    cmdclass={'install': InstallPlatlib},
    distclass=BinaryDistribution
)



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2007-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    plot_trajectories.py
# @author  Jakob Erdmann
# @date    2018-08-18

"""
This script plots fcd data for each vehicle using either
- distance vs speed
- time vs speed
- time vs distance

Individual trajectories can be clicked in interactive mode to print the vehicle Id on the console
"""
from __future__ import absolute_import
from __future__ import print_function
import os
import sys
from collections import defaultdict
from optparse import OptionParser
import matplotlib
if 'matplotlib.backends' not in sys.modules:
    if 'TEXTTEST_SANDBOX' in os.environ or (os.name == 'posix' and 'DISPLAY' not in os.environ):
        matplotlib.use('Agg')
import matplotlib.pyplot as plt  # noqa
import math  # noqa

from sumolib.xml import parse_fast_nested  # noqa
from sumolib.miscutils import uMin, uMax, parseTime  # noqa


def getOptions(args=None):
    optParser = OptionParser()
    optParser.add_option("-t", "--trajectory-type", dest="ttype", default="ds",
                         help="select two letters from [t, s, d, a, i, x, y] to plot"
                         + " Time, Speed, Distance, Acceleration, Angle, x-Position, y-Position."
                         + " Default 'ds' plots Distance vs. Speed")
    optParser.add_option("--persons", action="store_true", default=False, help="plot person trajectories")
    optParser.add_option("-s", "--show", action="store_true", default=False, help="show plot directly")
    optParser.add_option("-o", "--output", help="outputfile for saving plots", default="plot.png")
    optParser.add_option("--csv-output", dest="csv_output", help="write plot as csv", metavar="FILE")
    optParser.add_option("-b", "--ballistic", action="store_true", default=False,
                         help="perform ballistic integration of distance")
    optParser.add_option("--filter-route", dest="filterRoute",
                         help="only export trajectories that pass the given list of edges (regardless of gaps)")
    optParser.add_option("--filter-edges", dest="filterEdges",
                         help="only consider data for the given list of edges")
    optParser.add_option("--filter-ids", dest="filterIDs",
                         help="only consider data for the given list of vehicle (or person) ids")
    optParser.add_option("-p", "--pick-distance", dest="pickDist", type="float", default=1,
                         help="pick lines within the given distance in interactive plot mode")
    optParser.add_option("-i", "--invert-distance-angle", dest="invertDistanceAngle", type="float",
                         help="invert distance for trajectories with a average angle near FLOAT")
    optParser.add_option("--label", help="plot label (default input file name")
    optParser.add_option("--invert-yaxis", dest="invertYAxis", action="store_true",
                         default=False, help="Invert the Y-Axis")
    optParser.add_option("--legend", action="store_true", default=False, help="Add legend")
    optParser.add_option("-v", "--verbose", action="store_true", default=False, help="tell me what you are doing")

    options, args = optParser.parse_args(args=args)
    if len(args) < 1:
        sys.exit("mandatory argument FCD_FILE missing")
    options.fcdfiles = args

    if options.filterRoute is not None:
        options.filterRoute = options.filterRoute.split(',')
    if options.filterEdges is not None:
        options.filterEdges = set(options.filterEdges.split(','))
    if options.filterIDs is not None:
        options.filterIDs = set(options.filterIDs.split(','))
    return options


def write_csv(data, fname):
    with open(fname, 'w') as f:
        for veh, vals in sorted(data.items()):
            f.write('"%s"\n' % veh)
            for x in zip(*vals):
                f.write(" ".join(map(str, x)) + "\n")
            f.write('\n')


def short_names(filenames):
    if len(filenames) == 1:
        return filenames
    reversedNames = [''.join(reversed(f)) for f in filenames]
    prefixLen = len(os.path.commonprefix(filenames))
    suffixLen = len(os.path.commonprefix(reversedNames))
    return [f[prefixLen:-suffixLen] for f in filenames]


def onpick(event):
    mevent = event.mouseevent
    print("veh=%s x=%d y=%d" % (event.label, mevent.xdata, mevent.ydata))


def main(options):
    fig = plt.figure(figsize=(14, 9), dpi=100)
    fig.canvas.mpl_connect('pick_event', onpick)

    xdata = 2
    ydata = 1
    typespec = {
        't': ('Time', 0),
        's': ('Speed', 1),
        'd': ('Distance', 2),
        'a': ('Acceleration', 3),
        'i': ('Angle', 4),
        'x': ('x-Position', 5),
        'y': ('y-Position', 6),
    }

    shortFileNames = short_names(options.fcdfiles)
    if (len(options.ttype) == 2
            and options.ttype[0] in typespec
            and options.ttype[1] in typespec):
        xLabel, xdata = typespec[options.ttype[0]]
        yLabel, ydata = typespec[options.ttype[1]]
        plt.xlabel(xLabel)
        plt.ylabel(yLabel)
        plt.title(','.join(shortFileNames) if options.label is None else options.label)
    else:
        sys.exit("unsupported plot type '%s'" % options.ttype)

    element = 'vehicle'
    location = 'lane'
    if options.persons:
        element = 'person'
        location = 'edge'

    routes = defaultdict(list)  # vehID -> recorded edges
    # vehID -> (times, speeds, distances, accelerations, angles, xPositions, yPositions)
    data = defaultdict(lambda: ([], [], [], [], [], [], []))
    for fileIndex, fcdfile in enumerate(options.fcdfiles):
        for timestep, vehicle in parse_fast_nested(fcdfile, 'timestep', ['time'],
                                                   element, ['id', 'x', 'y', 'angle', 'speed', location]):
            vehID = vehicle.id
            if options.filterIDs and vehID not in options.filterIDs:
                continue
            if len(options.fcdfiles) > 1:
                suffix = shortFileNames[fileIndex]
                if len(suffix) > 0:
                    vehID += "#" + suffix
            if options.persons:
                edge = vehicle.edge
            else:
                edge = vehicle.lane[0:vehicle.lane.rfind('_')]
            if len(routes[vehID]) == 0 or routes[vehID][-1] != edge:
                routes[vehID].append(edge)
            if options.filterEdges and edge not in options.filterEdges:
                continue
            time = parseTime(timestep.time)
            speed = float(vehicle.speed)
            prevTime = time
            prevSpeed = speed
            prevDist = 0
            if vehID in data:
                prevTime = data[vehID][0][-1]
                prevSpeed = data[vehID][1][-1]
                prevDist = data[vehID][2][-1]
            data[vehID][0].append(time)
            data[vehID][1].append(speed)
            data[vehID][4].append(float(vehicle.angle))
            data[vehID][5].append(float(vehicle.x))
            data[vehID][6].append(float(vehicle.y))
            if prevTime == time:
                data[vehID][3].append(0)
            else:
                data[vehID][3].append((speed - prevSpeed) / (time - prevTime))

            if options.ballistic:
                avgSpeed = (speed + prevSpeed) / 2
            else:
                avgSpeed = speed
            data[vehID][2].append(prevDist + (time - prevTime) * avgSpeed)

    def line_picker(line, mouseevent):
        if mouseevent.xdata is None:
            return False, dict()
        # minxy = None
        # mindist = 10000
        for x, y in zip(line.get_xdata(), line.get_ydata()):
            dist = math.sqrt((x - mouseevent.xdata) ** 2 + (y - mouseevent.ydata) ** 2)
            if dist < options.pickDist:
                return True, dict(label=line.get_label())
            # else:
            #    if dist < mindist:
            #        print("   ", x,y, dist, (x - mouseevent.xdata) ** 2, (y - mouseevent.ydata) ** 2)
            #        mindist = dist
            #        minxy = (x, y)
        # print(mouseevent.xdata, mouseevent.ydata, minxy, dist,
        #        line.get_label())
        return False, dict()

    minY = uMax
    maxY = uMin
    minX = uMax
    maxX = uMin

    for vehID, d in data.items():
        if options.filterRoute is not None:
            skip = False
            route = routes[vehID]
            for required in options.filterRoute:
                if required not in route:
                    skip = True
                    break
            if skip:
                continue
        if options.invertDistanceAngle is not None:
            avgAngle = sum(d[4]) / len(d[4])
            if abs(avgAngle - options.invertDistanceAngle) < 45:
                maxDist = d[2][-1]
                for i, v in enumerate(d[2]):
                    d[2][i] = maxDist - v

        minY = min(minY, min(d[ydata]))
        maxY = max(maxY, max(d[ydata]))
        minX = min(minX, min(d[xdata]))
        maxX = max(maxX, max(d[xdata]))

        plt.plot(d[xdata], d[ydata], picker=line_picker, label=vehID)
    if options.invertYAxis:
        plt.axis([minX, maxX, maxY, minY])

    if options.legend > 0:
        plt.legend()

    plt.savefig(options.output)
    if options.csv_output is not None:
        write_csv(data, options.csv_output)
    if options.show:
        plt.show()


if __name__ == "__main__":
    main(getOptions())



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2010-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    matsim_importPlans.py
# @author  Jakob Erdmann
# @author  Camillo Fillinger
# @date    2019-09-27

"""
Import person plans from MATSim
"""
from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import optparse
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
import sumolib  # noqa


def get_options(args=None):
    optParser = optparse.OptionParser()
    optParser.add_option("-p", "--plan-file", dest="planfile",
                         help="define the route file (mandatory)")
    optParser.add_option("-o", "--out-file", dest="outfile",
                         help="Output file (mandatory)")
    optParser.add_option("--vehicles-only", dest="carsOnly", action="store_true",
                         default=False, help="Import only vehicles instead of persons")
    optParser.add_option("--default-start", dest="defaultStart", default="0:0:0",
                         help="default start time for the first activity")
    optParser.add_option("--default-end", dest="defaultEnd", default="24:0:0",
                         help="default end time for the last activity")
    optParser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                         default=False, help="tell me what you are doing")

    (options, args) = optParser.parse_args(args=args)

    if not options.planfile or not options.outfile:
        optParser.print_help()
        sys.exit()

    return options


def writeLeg(outf, options, idveh, leg, startLink, endLink):
    depart = leg.dep_time if options.carsOnly else "triggered"
    if leg.route is None or leg.route[0].distance == "NaN":
        outf.write('   <trip id="%s" depart="%s" from="%s" to="%s"/>\n'
                   % (idveh, depart, startLink, endLink))
    else:
        outf.write('   <vehicle id="%s" depart="%s" >\n' % (idveh, depart))
        outf.write('        <route edges="%s"/>\n' % (leg.route[0].getText()))
        outf.write('   </vehicle>\n')


def main(options):
    persons = []  # (depart, xmlsnippet)
    for person in sumolib.xml.parse(options.planfile, 'person'):
        outf = StringIO()
        vehIndex = 0
        plan = person.plan[0]
        if len(plan.getChildList()) == 0:
            continue
        firstAct = plan.getChildList()[0]
        depart = firstAct.start_time
        if depart is None:
            depart = options.defaultStart
        attributes = person.attributes[0] if person.attributes else None
        # write vehicles
        vehicleslist = []
        untillist = []
        lastAct = None
        lastLeg = None
        for item in plan.getChildList():
            leg = None
            idveh = "%s_%s" % (person.id, vehIndex)
            if "act" in item.name:  # act or activity
                if lastLeg is not None:
                    leg = lastLeg
                    leg.dep_time = lastAct.end_time
                    writeLeg(outf, options, idveh, leg, lastAct.link, item.link)
                    lastLeg = None
                lastAct = item
            if item.name == "leg":
                if item.route is None:
                    lastLeg = item
                else:
                    leg = item
                    writeLeg(outf, options, idveh, leg, leg.route[0].start_link, leg.route[0].end_link)
            if leg:
                untillist.append(leg.dep_time)
                vehicleslist.append(idveh)
                vehIndex = vehIndex+1
        untillist.append(lastAct.end_time if lastAct.end_time else options.defaultEnd)
        # write person
        if not options.carsOnly:
            vehIndex = 0
            outf.write('   <person id="%s" depart="%s">\n' % (person.id, depart))
            if attributes is not None:
                for attr in attributes.attribute:
                    outf.write('       <param key="%s" value="%s"/>\n' % (attr.attr_name, attr.getText()))

            lastLeg = None
            for item in plan.getChildList():
                if "act" in item.name:  # act or activity
                    if lastLeg is not None:
                        outf.write('       <ride lines="%s" to="%s"  />\n' % (vehicleslist[vehIndex], item.link))
                        vehIndex = vehIndex+1
                    outf.write('       <stop lane="%s_0" until="%s" actType="%s" />\n' %
                               (item.link, untillist[vehIndex], item.type))
                if item.name == "leg":
                    lastLeg = item
            outf.write('   </person>\n')
        persons.append((sumolib.miscutils.parseTime(depart), outf.getvalue()))

    persons.sort()
    with open(options.outfile, 'w') as outf:
        sumolib.writeXMLHeader(outf, "$Id: matsim_importPlans.py v1_3_1+0713-63b241ac79 michael.behrisch@dlr.de 2019-10-20 15:41:56 +0200 $", "routes")  # noqa
        for depart, xml in persons:
            outf.write(xml)
        outf.write('</routes>\n')
    outf.close()


if __name__ == "__main__":
    options = get_options(sys.argv)
    main(options)



# -*- coding: utf-8 -*-
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2008-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    _variablespeedsign.py
# @author  Matthias Schwamborn
# @date    2020-08-18

from __future__ import absolute_import
from . import constants as tc
from .domain import Domain


class VariableSpeedSignDomain(Domain):

    def __init__(self):
        Domain.__init__(self, "variablespeedsign",
                        tc.CMD_GET_VARIABLESPEEDSIGN_VARIABLE, tc.CMD_SET_VARIABLESPEEDSIGN_VARIABLE,
                        tc.CMD_SUBSCRIBE_VARIABLESPEEDSIGN_VARIABLE, tc.RESPONSE_SUBSCRIBE_VARIABLESPEEDSIGN_VARIABLE,
                        tc.CMD_SUBSCRIBE_VARIABLESPEEDSIGN_CONTEXT, tc.RESPONSE_SUBSCRIBE_VARIABLESPEEDSIGN_CONTEXT)

    def getLanes(self, vssID):
        """getLanes(string) -> list(string)

        Returns a list of all lanes controlled by the variablespeedsign.
        """
        return self._getUniversal(tc.VAR_LANES, vssID)



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    nefz.py
# @author  daniel.krajzewicz@dlr.de
# @date    2014-01-14

"""
Generates a ';'-separated file that contains the time line of the NEFZ
 driving cycle.
"""
from __future__ import print_function

NEFZ1 = [
    [11, 0, 0],
    [4, 1.04, -1], [0, 0, 15],
    [8, 0, 15],
    [2, -0.69, -1], [0, 0, 10],
    [3, -.93, -1], [0, 0, 0],
    [21, 0, 0],

    [5, 0.83, -1], [0, 0, 15],
    [2, 0, -1],
    [5, 0.94, -1], [0, 0, 32],
    [24, 0, 32],
    [8, -.76, -1], [0, 0, 10],
    [3, -.93, -1], [0, 0, 0],
    [21, 0, 0],

    [5, .83, -1], [0, 0, 15],
    [2, 0, -1],
    [9, .62, -1], [0, 0, 35],
    [2, 0, -1],
    [8, .52, -1], [0, 0, 50],
    [12, 0, 50],
    [8, -.52, -1], [0, 0, 35],
    [13, 0, 35],
    [2, 0, -1],
    [7, -.87, -1], [0, 0, 10],
    [3, -.93, -1], [0, 0, 0],

    [7, 0, 0],

]

NEFZ2 = [
    [20, 0, 0],
    [5, .83, -1], [0, 0, 15],
    [2, 0, -1],
    [9, .62, -1], [0, 0, 35],
    [2, 0, -1],
    [8, .52, -1], [0, 0, 50],
    [2, 0, -1],
    [13, .43, -1], [0, 0, 70],
    [50, 0, 70],
    [8, -.69, -1], [0, 0, 50],
    [69, 0, 50],
    [13, .43, -1], [0, 0, 70],
    [50, 0, 70],
    [35, .24, -1], [0, 0, 100],
    [30, 0, 100],
    [20, .28, -1], [0, 0, 120],
    [10, 0, 120],
    [16, -.69, -1], [0, 0, 80],
    [8, -1.04, -1], [0, 0, 50],
    [10, -1.39, -1], [0, 0, 0],
    [20, 0, 0]
]


def build(what):
    t = 0
    v = 0
    a = 0
    ts1 = []
    vs1 = []
    as1 = []
    ts2 = []
    vs2 = []
    as2 = []
    ts3 = []
    vs3 = []
    as3 = []
    ct = 0
    cv = 0
    lv = 0
    lt = 0
    ts1.append(0)
    as1.append(0)
    vs1.append(0)
    for tav in what:
        [t, a, v] = tav[:3]
        v = v / 3.6
        if v >= 0:
            # destination velocity
            if a != 0:
                print("ups %s" % tav)
            ts1.append(ct + t)
            as1.append(0)
            vs1.append(v)
        # via acceleration
        for it in range(0, t):
            ts2.append(ct + it)
            as2.append(a)
            mv = cv + a * float(it)
            if mv < 0:
                mv = 0
            vs2.append(mv)
        # via speed (if not None, otherwise "keep in mind")
        if v >= 0:
            dt = float((ct + t) - lt)
            if dt != 0:
                dv = float(v - lv)
                a = dv / float(dt)
                for it in range(lt, ct + t):
                    ts3.append(it)
                    as3.append(a)
                    vs3.append(lv + a * float(it - lt))

        ct = ct + t
        if v >= 0:
            cv = v
            lv = v
            lt = ct
    return [ts1, vs1, as1, ts2, vs2, as2, ts3, vs3, as3]


BASE = 3
ts = []
vs = []
ts1 = []
vs1 = []
ts2 = []
vs2 = []
ts3 = []
vs3 = []
t = 0
for c in [NEFZ1, NEFZ1, NEFZ1, NEFZ1, NEFZ2]:
    tmp = build(c)
    for i in range(0, len(tmp[BASE])):
        ts.append(tmp[BASE][i] + t)
        vs.append(tmp[BASE + 1][i])
        # ts1.append(tmp[0][i]+t)
        # vs1.append(tmp[0+1][i])
        # ts2.append(tmp[3][i]+t)
        # vs2.append(tmp[3+1][i])
        # ts3.append(tmp[6][i]+t)
        # vs3.append(tmp[6+1][i])
    t = t + tmp[BASE][-1] + 1

fdo = open("nefz.csv", "w")
pv = 0
for i in range(0, len(ts)):
    fdo.write("%s;%s;%s\n" % (ts[i], vs[i] * 3.6, pv - vs[i]))
    pv = vs[i]
fdo.close()



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    generateRailSignalConstraints.py
# @author  Jakob Erdmann
# @date    2020-08-31

"""
Generate railSignalConstrains definitions that enforce a loaded rail schedule

Two types of constraints are generated:
1. <predecessor>
When two vehices stop subsequently at the same busStop (trainStop) and they reach that stop
via different routes, the switch where both routes merge is identified and a
constraint is created for the rail signals that guard this merging switch:
    The vehicle B that arrives at the stop later, must wait (at its signal Y)
    for the vehicle A that arrives first (to pass it's respective signal X)
    This uses the 'arrival' attribute of the vehicle stops

A complication arrises if the signal of the first vehicle is passed by other
trains which are en route to another stop. This makes it necessary to record a
larger number of passing vehicles within the simulation (controlled by the
limit attribute). The script attempts to determine the necessary limit value by
identifying all vehicles that pass the signal X en route to other stops between
the time A and B reach their respective signals (counting backwards from the
next stop based on "arrival". To account for delays the
options --delay and --limit can be used to override the limit values

2. <insertionPredecessor>
Whenever a vehicle B departs at a stop (assumed to coincide with the "until"
attribute of it's first stop), the prior train A that leaves this stop is
identified (also based on "until"). Then a constraint is created that prevents
insertion of B until train A has passed the next signal that lies beyond the
stop.
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import subprocess
from collections import defaultdict
from operator import itemgetter

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
import sumolib  # noqa
from sumolib.miscutils import parseTime, parseBool, humanReadableTime  # noqa

DUAROUTER = sumolib.checkBinary('duarouter')


def get_options(args=None):
    parser = sumolib.options.ArgumentParser(description="Sample routes to match counts")
    parser.add_argument("-n", "--net-file", dest="netFile",
                        help="Input network file")
    parser.add_argument("-a", "--additional-file", dest="addFile",
                        help="Input additional file (busStops)")
    parser.add_argument("-t", "--trip-file", dest="tripFile",
                        help="Input trip file (will be processed into a route file)")
    parser.add_argument("-r", "--route-file", dest="routeFile",
                        help="Input route file (must contain routed vehicles rather than trips)")
    parser.add_argument("-o", "--output-file", dest="out", default="constraints.add.xml",
                        help="Output additional file")
    parser.add_argument("-b", "--begin", default="0",
                        help="ignore vehicles departing before the given begin time (seconds or H:M:S)")
    parser.add_argument("--until-from-duration", action="store_true", default=False, dest="untilFromDuration",
                        help="Use stop arrival+duration instead of 'until' to compute insertion constraints")
    parser.add_argument("-d", "--delay", default="0",
                        help="Assume given maximum delay when computing the number of intermediate vehicles " +
                        "that pass a given signal (for setting limit)")
    parser.add_argument("-l", "--limit", type=int, default=0,
                        help="Increases the limit value for tracking passed vehicles by the given amount")
    parser.add_argument("--abort-unordered", dest="abortUnordered", action="store_true", default=False,
                        help="Abort generation of constraints for a stop "
                        "once the ordering of vehicles by 'arrival' differs from the ordering by 'until'")
    parser.add_argument("-p", "--ignore-parking", dest="ignoreParking", action="store_true", default=False,
                        help="Ignore unordered timing if the vehicle which arrives first is parking")
    parser.add_argument("-P", "--skip-parking", dest="skipParking", action="store_true", default=False,
                        help="Do not generate constraints for a vehicle that parks at the next stop")
    parser.add_argument("--comment.line", action="store_true", dest="commentLine", default=False,
                        help="add lines of involved trains in comment")
    parser.add_argument("--comment.id", action="store_true", dest="commentId", default=False,
                        help="add ids of involved trains in comment (when different from tripId)")
    parser.add_argument("-v", "--verbose", action="store_true", default=False,
                        help="tell me what you are doing")
    parser.add_argument("--debug-switch", dest="debugSwitch",
                        help="print debug information for the given merge-switch edge")
    parser.add_argument("--debug-signal", dest="debugSignal",
                        help="print debug information for the given signal id")
    parser.add_argument("--debug-stop", dest="debugStop",
                        help="print debug information for the given busStop id")

    options = parser.parse_args(args=args)
    if (options.routeFile is None and options.tripFile is None) or options.netFile is None:
        parser.print_help()
        sys.exit()

    if options.routeFile is None:
        options.routeFile = options.tripFile + ".rou.xml"
        args = [DUAROUTER, '-n', options.netFile,
                '-r', options.tripFile,
                '-a', options.addFile,
                '-o', options.routeFile,
                '--ignore-errors', '--no-step-log',
                ]
        print("calling", " ".join(args))
        sys.stdout.flush()
        subprocess.call(args)
        sys.stdout.flush()

    options.delay = parseTime(options.delay)

    return options


def getTravelTime(net, edges):
    result = 0
    for edgeID in edges:
        edge = net.getEdge(edgeID)
        result += edge.getLength() / edge.getSpeed()
    return result


def getStopEdges(addFile):
    """find edge for each stopping place"""
    stopEdges = {}
    for busstop in sumolib.xml.parse(addFile, 'busStop'):
        stopEdges[busstop.id] = sumolib._laneID2edgeID(busstop.lane)

    print("read %s busStops on %s edges" % (len(stopEdges), len(set(stopEdges.values()))))
    return stopEdges


def getStopRoutes(options, stopEdges):
    """parse routes and determine the list of edges between stops
        return: setOfUniqueRoutes, busstopDict
    """
    uniqueRoutes = set()
    stopRoutes = defaultdict(list)  # busStop -> [(edges, stopObj), ....]
    vehicleStopRoutes = defaultdict(list)  # vehID -> [(edges, stopObj), ....]
    numRoutes = 0
    numStops = 0
    begin = parseTime(options.begin)
    for vehicle in sumolib.xml.parse(options.routeFile, 'vehicle', heterogeneous=True):
        depart = parseTime(vehicle.depart)
        if depart < begin:
            continue
        numRoutes += 1
        edges = tuple(vehicle.route[0].edges.split())
        uniqueRoutes.add(edges)
        lastIndex = -1
        routeIndex = 0
        tripId = vehicle.id
        line = vehicle.getAttributeSecure("line", "")
        for stop in vehicle.stop:
            numStops += 1
            if stop.busStop is None:
                stop.setAttribute("busStop", stop.lane)
                stopEdges[stop.lane] = sumolib._laneID2edgeID(stop.lane)
            stopEdge = stopEdges[stop.busStop]
            while edges[routeIndex] != stopEdge:
                routeIndex += 1
                assert(routeIndex < len(edges))
            edgesBefore = edges[lastIndex + 1: routeIndex + 1]
            stop.setAttribute("prevTripId", tripId)
            stop.setAttribute("prevLine", line)
            stop.setAttribute("vehID", vehicle.id)
            tripId = stop.getAttributeSecure("tripId", tripId)
            line = stop.getAttributeSecure("line", line)
            stopRoutes[stop.busStop].append((edgesBefore, stop))
            vehicleStopRoutes[vehicle.id].append((edgesBefore, stop))
            lastIndex = routeIndex

    print("read %s routes (%s unique) and %s stops at %s busStops" % (
        numRoutes, len(uniqueRoutes), numStops, len(stopRoutes)))

    return uniqueRoutes, stopRoutes, vehicleStopRoutes


def findMergingSwitches(options, uniqueRoutes, net):
    """find switches where routes merge and thus conflicts must be solved"""
    predEdges = defaultdict(set)
    predReversal = set()
    for edges in uniqueRoutes:
        for i, edge in enumerate(edges):
            if i > 0:
                pred = edges[i - 1]
                if net.getEdge(edge).getBidi() == net.getEdge(pred):
                    predReversal.add(edge)
                predEdges[edge].add(pred)

    mergeSwitches = set()
    numReversals = 0
    for edge, preds in predEdges.items():
        if len(preds) > 1:
            if edge in predReversal:
                numReversals += 1
            if options.verbose:
                print("mergingEdge=%s pred=%s" % (edge, ','.join(preds)))
            mergeSwitches.add(edge)

    if numReversals == 0:
        reversalInfo = ""
    else:
        reversalInfo = " (including %s reversal-merges)" % numReversals
    print("processed %s routes across %s edges with %s merging switches%s" % (
        len(uniqueRoutes), len(predEdges), len(mergeSwitches), reversalInfo))
    return mergeSwitches


def findStopsAfterMerge(net, stopRoutes, mergeSwitches):
    """find stops at the same busStop that come directly after a the same merge switch (no prior step before
    the switch). Returns filtered stopRoutes"""
    switchRoutes = defaultdict(lambda: defaultdict(list))  # mergeSwitch -> busStop -> [(edges, stopObj), ....]
    mergeSignals = {}  # (switch, edges) -> signal
    numFound = 0
    for busStop, stops in stopRoutes.items():
        for edgesBefore, stop in stops:
            signal = None
            signalEdgeIndex = 0
            prevEdge = None
            for i, edge in enumerate(edgesBefore):
                node = net.getEdge(edge).getFromNode()
                if node.getType() == "rail_signal":
                    tls = net.getTLS(node.getID())
                    for inLane, outLane, _ in tls.getConnections():
                        if (outLane.getEdge().getID() == edge
                                and (prevEdge is None or prevEdge == inLane.getEdge().getID())):
                            signal = tls.getID()
                            signalEdgeIndex = i
                            break
                if edge in mergeSwitches:
                    numFound += 1
                    switchRoutes[edge][busStop].append((edgesBefore, stop))
                    if signal is None:
                        print("No signal found when approaching stop %s via switch %s along route %s" % (
                            busStop, edge, ','.join(edgesBefore)), file=sys.stderr)
                    # travel time from signal to stop
                    ttSignalStop = getTravelTime(net, edgesBefore[signalEdgeIndex:])
                    mergeSignals[(edge, edgesBefore)] = (signal, ttSignalStop)
                prevEdge = edge

    print("Found %s stops after merging switches and %s signals that guard switches" % (
        numFound, len(set(mergeSignals.values()))))
    return switchRoutes, mergeSignals


def computeSignalTimes(options, net, stopRoutes):
    signalTimes = defaultdict(list)  # signal -> [(timeAtSignal, stop), ...]
    debugInfo = []
    for _, stops in stopRoutes.items():
        for edgesBefore, stop in stops:
            if stop.hasAttribute("arrival"):
                arrival = parseTime(stop.arrival)
            elif stop.hasAttribute("until"):
                arrival = parseTime(stop.until) - parseTime(stop.getAttributeSecure("duration", "0"))
            else:
                continue
            for i, edge in enumerate(edgesBefore):
                node = net.getEdge(edge).getFromNode()
                if node.getType() == "rail_signal":
                    tls = net.getTLS(node.getID())
                    for _, outLane, __ in tls.getConnections():
                        if outLane.getEdge().getID() == edge:
                            signal = tls.getID()
                            ttSignalStop = getTravelTime(net, edgesBefore[i:])
                            timeAtSignal = arrival - ttSignalStop
                            signalTimes[signal].append((timeAtSignal, stop))
                            if signal == options.debugSignal:
                                debugInfo.append((timeAtSignal,
                                                  ("%s vehID=%s prevTripId=%s passes signal %s to stop %s " +
                                                   "arrival=%s ttSignalStop=%s edges=%s") % (
                                                      humanReadableTime(timeAtSignal),
                                                      stop.vehID, stop.prevTripId,
                                                      signal, stop.busStop,
                                                      humanReadableTime(arrival), ttSignalStop,
                                                      edgesBefore)))
                            break
    for signal in signalTimes.keys():
        signalTimes[signal] = sorted(signalTimes[signal])

    if options.debugSignal in signalTimes:
        for _, info in sorted(debugInfo):
            print(info)
        busStops = set([s.busStop for a, s in signalTimes[options.debugSignal]])
        arrivals = [a for a, s in signalTimes[options.debugSignal]]
        print("Signal %s is passed %s times between %s and %s on approach to stops %s" % (
            options.debugSignal, len(arrivals), arrivals[0], arrivals[-1], ' '.join(busStops)))

    return signalTimes


def countPassingTrainsToOtherStops(options, signal, currentStop, begin, end, signalTimes):
    count = 0
    for timeAtSwitch, stop in signalTimes[signal]:
        if timeAtSwitch < begin:
            continue
        if timeAtSwitch > end:
            break
        if stop.busStop == currentStop:
            continue
        count += 1
        if options.debugSignal == signal:
            print("vehicle %s (%s) passes signal %s at time %s (in between %s and %s)" % (
                stop.prevTripId, stop.vehID, signal, timeAtSwitch, begin, end))
    return count


def findConflicts(options, switchRoutes, mergeSignals, signalTimes):
    """find stops that target the same busStop from different branches of the
    prior merge switch and establish their ordering"""

    numConflicts = 0
    numIgnoredConflicts = 0
    # signal -> [(tripID, otherSignal, otherTripID, limit, line, otherLine, vehID, otherVehID), ...]
    conflicts = defaultdict(list)
    ignoredVehicles = set()
    for switch, stopRoutes2 in switchRoutes.items():
        numSwitchConflicts = 0
        numIgnoredSwitchConflicts = 0
        if switch == options.debugSwitch:
            print("Switch %s lies ahead of busStops %s" % (switch, stopRoutes2.keys()))
        for busStop, stops in stopRoutes2.items():
            arrivals = []
            for edges, stop in stops:
                if stop.hasAttribute("arrival"):
                    arrival = parseTime(stop.arrival)
                elif stop.hasAttribute("until"):
                    arrival = parseTime(stop.until) - parseTime(stop.getAttributeSecure("duration", "0"))
                else:
                    print("ignoring stop at %s without schedule information (arrival, until)" % busStop)
                    continue
                arrivals.append((arrival, edges, stop))
            arrivals.sort(key=itemgetter(0))
            ignore = False
            for (pArrival, pEdges, pStop), (nArrival, nEdges, nStop) in zip(arrivals[:-1], arrivals[1:]):
                pSignal, pTimeSiSt = mergeSignals[(switch, pEdges)]
                nSignal, nTimeSiSt = mergeSignals[(switch, nEdges)]
                if switch == options.debugSwitch:
                    print(pSignal, nSignal, pStop, nStop)
                if pSignal != nSignal and pSignal is not None and nSignal is not None:
                    if (ignore or (options.abortUnordered
                                   and pStop.hasAttribute("until")
                                   and nStop.hasAttribute("until")
                                   and (not options.ignoreParking or not
                                        parseBool(pStop.getAttributeSecure("parking", "false")))
                                   and parseTime(pStop.until) > parseTime(nStop.until))):
                        numIgnoredConflicts += 1
                        numIgnoredSwitchConflicts += 1
                        # ignore conflict and any that follow
                        if not ignore:
                            print("Found inconsistent times at stop %s "
                                  "for vehicle %s (%s, %s) and vehicle %s (%s, %s)" % (
                                      busStop,
                                      pStop.vehID,
                                      humanReadableTime(pArrival),
                                      humanReadableTime(parseTime(pStop.until)),
                                      nStop.vehID,
                                      humanReadableTime(nArrival),
                                      humanReadableTime(parseTime(nStop.until))),
                                  file=sys.stderr)
                            # ignoredVehicles.insert(pStop.vehID)
                            ignoredVehicles.add(nStop.vehID)
                        ignore = True
                        continue
                    if (options.abortUnordered and nStop.vehID in ignoredVehicles):
                        # no constraints for inconsistent vehicle to avoid deadlock
                        numIgnoredConflicts += 1
                        numIgnoredSwitchConflicts += 1
                        continue
                    if options.skipParking and parseBool(nStop.getAttributeSecure("parking", "false")):
                        print("ignoring stop at %s for parking vehicle %s (%s, %s)" % (
                            busStop, nStop.vehID, humanReadableTime(nArrival),
                            (humanReadableTime(parseTime(nStop.until)) if nStop.hasAttribute("until") else "-")))
                        numIgnoredConflicts += 1
                        numIgnoredSwitchConflicts += 1
                        continue
                    if options.skipParking and parseBool(pStop.getAttributeSecure("parking", "false")):
                        print("ignoring stop at %s for %s (%s, %s) after parking vehicle %s (%s, %s)" % (
                            busStop, nStop.vehID, humanReadableTime(nArrival),
                            (humanReadableTime(parseTime(nStop.until)) if nStop.hasAttribute("until") else "-"),
                            pStop.vehID, humanReadableTime(pArrival),
                            (humanReadableTime(parseTime(pStop.until)) if pStop.hasAttribute("until") else "-")))
                        numIgnoredConflicts += 1
                        numIgnoredSwitchConflicts += 1
                        continue

                    numConflicts += 1
                    numSwitchConflicts += 1
                    # check for trains that pass the switch in between the
                    # current two trains (heading to another stop) and raise the limit
                    limit = 1
                    pTimeAtSignal = pArrival - pTimeSiSt
                    nTimeAtSignal = nArrival - nTimeSiSt
                    end = nTimeAtSignal + options.delay
                    if options.verbose and options.debugSignal == pSignal:
                        print("check vehicles between %s and %s (including delay %s) at signal %s pStop=%s nStop=%s" % (
                            humanReadableTime(pTimeAtSignal),
                            humanReadableTime(end), options.delay, pSignal,
                            pStop, nStop))
                    limit += countPassingTrainsToOtherStops(options, pSignal, busStop, pTimeAtSignal, end, signalTimes)
                    conflicts[nSignal].append((nStop.prevTripId, pSignal, pStop.prevTripId, limit,
                                               # attributes for adding comments
                                               nStop.prevLine, pStop.prevLine, nStop.vehID, pStop.vehID))
        if options.verbose:
            print("Found %s conflicts at switch %s" % (numSwitchConflicts, switch))
            if numIgnoredSwitchConflicts > 0:
                print("Ignored %s conflicts at switch %s" % (numIgnoredSwitchConflicts, switch))

    print("Found %s conflicts" % numConflicts)
    if numIgnoredConflicts > 0:
        print("Ignored %s conflicts" % numIgnoredConflicts)
    return conflicts


def findSignal(net, nextEdges):
    prevEdge = None
    for edge in nextEdges:
        node = net.getEdge(edge).getFromNode()
        if node.getType() == "rail_signal":
            tls = net.getTLS(node.getID())
            for inLane, outLane, _ in tls.getConnections():
                if (outLane.getEdge().getID() == edge
                        and (prevEdge is None or prevEdge == inLane.getEdge().getID())):
                    return tls.getID()
        prevEdge = edge
    return None


def findInsertionConflicts(options, net, stopEdges, stopRoutes, vehicleStopRoutes):
    """find routes that start at a stop with a traffic light at end of the edge
    and routes that pass this stop. Ensure
    insertion happens in the correct order"""
    # signal -> [(tripID, otherSignal, otherTripID, limit, line, otherLine, vehID, otherVehID), ...]
    conflicts = defaultdict(list)
    numConflicts = 0
    numIgnoredConflicts = 0
    for busStop, stops in stopRoutes.items():
        stopEdge = stopEdges[busStop]
        node = net.getEdge(stopEdge).getToNode()
        signal = node.getID()
        untils = []
        for edgesBefore, stop in stops:
            if stop.hasAttribute("until") and not options.untilFromDuration:
                until = parseTime(stop.until)
            elif stop.hasAttribute("arrival"):
                until = parseTime(stop.arrival) + parseTime(stop.getAttributeSecure("duration", "0"))
            else:
                continue
            untils.append((until, edgesBefore, stop))
        # only use 'until' for sorting and keep the result stable otherwise
        untils.sort(key=itemgetter(0))
        prevPassing = None
        ignore = False
        for i, (nUntil, nEdges, nStop) in enumerate(untils):
            nVehStops = vehicleStopRoutes[nStop.vehID]
            nIndex = nVehStops.index((nEdges, nStop))
            nIsPassing = nIndex < len(nVehStops) - 1
            nIsDepart = len(nEdges) == 1
            if options.verbose and busStop == options.debugStop:
                print(i,
                      "n:", humanReadableTime(nUntil), nStop.tripId, nStop.vehID, nIndex, len(nVehStops),
                      "passing:", nIsPassing,
                      "depart:", nIsDepart)
            if prevPassing is not None and nIsDepart:
                pUntil, pEdges, pStop = prevPassing
                pVehStops = vehicleStopRoutes[pStop.vehID]
                pIndex = pVehStops.index((pEdges, pStop))
                # no need to constrain subsequent departures (simulation should maintain ordering)
                if len(pEdges) > 1 or pIndex > 0:
                    # find edges after stop
                    if busStop == options.debugStop:
                        print(i,
                              "p:", humanReadableTime(pUntil), pStop.tripId, pStop.vehID, pIndex, len(pVehStops),
                              "n:", humanReadableTime(nUntil), nStop.tripId, nStop.vehID, nIndex, len(nVehStops))
                    if nIsPassing:
                        # both vehicles move past the stop
                        pNextEdges = pVehStops[pIndex + 1][0]
                        nNextEdges = nVehStops[nIndex + 1][0]
                        limit = 1  # recheck
                        pSignal = signal
                        nSignal = signal
                        if node.getType() != "rail_signal":
                            # find signal in nextEdges
                            pSignal = findSignal(net, pNextEdges)
                            nSignal = findSignal(net, nNextEdges)
                            if pSignal is None or nSignal is None:
                                print(("Ignoring insertion conflict between %s and %s at stop '%s' " +
                                       "because no rail signal was found after the stop") % (
                                    nStop.prevTripId, pStop.prevTripId, busStop), file=sys.stderr)
                                continue
                        # check for inconsistent ordering
                        if (ignore or (options.abortUnordered
                                       and pStop.hasAttribute("arrival")
                                       and nStop.hasAttribute("arrival")
                                       and (not options.ignoreParking or not
                                            parseBool(nStop.getAttributeSecure("parking", "false")))
                                       and parseTime(pStop.arrival) > parseTime(nStop.arrival))):
                            numIgnoredConflicts += 1
                            # ignore conflict and any that follow
                            if not ignore:
                                # sort output by arrival again
                                print("Found inconsistent times at stop %s "
                                      "for vehicle %s (%s, %s) and vehicle %s (%s, %s)" % (
                                          busStop,
                                          nStop.vehID,
                                          humanReadableTime(parseTime(nStop.arrival)),
                                          humanReadableTime(nUntil),
                                          pStop.vehID,
                                          humanReadableTime(parseTime(pStop.arrival)),
                                          humanReadableTime(pUntil)),
                                      file=sys.stderr)
                            ignore = True
                            continue
                        # predecessor tripId after stop is needed
                        pTripId = pStop.getAttributeSecure("tripId", pStop.vehID)
                        conflicts[nSignal].append((nStop.prevTripId, pSignal, pTripId, limit,
                                                   # attributes for adding comments
                                                   nStop.prevLine, pStop.prevLine, nStop.vehID, pStop.vehID))
                        numConflicts += 1
                        if busStop == options.debugStop:
                            print("   found insertionConflict pSignal=%s nSignal=%s pTripId=%s" % (
                                pSignal, nSignal, pTripId)),

            if nIsPassing:
                prevPassing = (nUntil, nEdges, nStop)

    print("Found %s insertion conflicts" % numConflicts)
    if numIgnoredConflicts > 0:
        print("Ignored %s insertion conflicts" % (numIgnoredConflicts))
    return conflicts


def writeConstraint(options, outf, tag, values):
    tripID, otherSignal, otherTripID, limit, line, otherLine, vehID, otherVehID = values
    comment = ""
    if options.commentLine:
        if line != "":
            comment += "line=%s " % line
        if otherLine != "":
            comment += "foeLine=%s " % otherLine
    if options.commentId:
        if vehID != tripID:
            comment += "vehID=%s " % vehID
        if otherVehID != otherTripID:
            comment += "foeID=%s " % otherVehID
    limit += options.limit
    if comment != "":
        comment = "   <!-- %s -->" % comment
    if limit == 1:
        limit = ""
    else:
        limit = ' limit="%s"' % limit
    outf.write('        <%s tripId="%s" tl="%s" foes="%s"%s/>%s\n' % (
        tag, tripID, otherSignal, otherTripID, limit, comment))


def main(options):
    net = sumolib.net.readNet(options.netFile)
    stopEdges = getStopEdges(options.addFile)
    uniqueRoutes, stopRoutes, vehicleStopRoutes = getStopRoutes(options, stopEdges)
    mergeSwitches = findMergingSwitches(options, uniqueRoutes, net)
    signalTimes = computeSignalTimes(options, net, stopRoutes)
    switchRoutes, mergeSignals = findStopsAfterMerge(net, stopRoutes, mergeSwitches)
    conflicts = findConflicts(options, switchRoutes, mergeSignals, signalTimes)
    insertionConflicts = findInsertionConflicts(options, net, stopEdges, stopRoutes, vehicleStopRoutes)

    signals = sorted(set(list(conflicts.keys()) + list(insertionConflicts.keys())))
    with open(options.out, "w") as outf:
        sumolib.writeXMLHeader(outf, "$Id$", "additional")  # noqa
        for signal in signals:
            outf.write('    <railSignalConstraints id="%s">\n' % signal)
            for values in conflicts[signal]:
                writeConstraint(options, outf, "predecessor", values)
            for values in insertionConflicts[signal]:
                writeConstraint(options, outf, "insertionPredecessor", values)
            outf.write('    </railSignalConstraints>\n')
        outf.write('</additional>\n')


if __name__ == "__main__":
    main(get_options())



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2010-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    findAllRoutes.py
# @author  Michael Behrisch
# @date    2020-11-02

"""
determine all possible routes between source and destination edges
"""
from __future__ import absolute_import
from __future__ import print_function

import sys
import sumolib


def get_options(args=None):
    argParser = sumolib.options.ArgumentParser()
    argParser.add_argument("-n", "--net-file", help="the SUMO net filename")
    argParser.add_argument("-o", "--output-file", help="the route output filename")
    argParser.add_argument("-s", "--source-edges", help="comma separated list of source edge ids")
    argParser.add_argument("-t", "--target-edges", help="comma separated list of target edge ids")
    options = argParser.parse_args(args=args)
    if options.net_file is None or options.output_file is None:
        argParser.print_help()
        sys.exit()
    return options


def main(options):
    net = sumolib.net.readNet(options.net_file)
    if options.source_edges is None:
        stubs = [[e] for e in net.getEdges()]
    else:
        stubs = [[net.getEdge(e)] for e in options.source_edges.split(",")]
    if options.target_edges is None:
        targets = set(net.getEdges())
    else:
        targets = set([net.getEdge(e) for e in options.target_edges.split(",")])
    with open(options.output_file, 'w') as outf:
        sumolib.xml.writeHeader(outf, root="routes")
        idx = 0
        while stubs:
            s = stubs.pop(0)
            if s[-1] in targets:
                print('    <route id="%s" edges="%s"/>' % (idx, " ".join([e.getID() for e in s])), file=outf)
                idx += 1
            else:
                seen = set(s)
                for edge in sorted(s[-1].getOutgoing()):
                    if edge not in seen:
                        stubs.append(s + [edge])
        outf.write('</routes>\n')


if __name__ == "__main__":
    main(get_options())
